(* Mathematica package *)

(* Performs some Tests for Quantum Object structure *)

BeginPackage["QuantumAlgebra`Analyze`", { "QuantumAlgebra`", "QuantumAlgebra`Options`" }]


Analyze::usage = "Prints out some disgnostic messages about Analyze result."

QuantumFreeQ::usage = "Tests if expression is free of Quantum objects."

QuantumExpressionQ::usage = "Tests if expression is well formed Quantum expression."

BraExpressionQ::usage = "Tests for Bra."

KetExpressionQ::usage = "Tests for Ket."

OperatorExpressionQ::usage = "Tests for Quantum Operator."

CommutatorExpressionQ::usage = "Tests for Quantum Commutator."

ConstantExpressionQ::usage = "Tests for Constant."

QuantumConstantExpressionQ::usage = "Tests for Quantum Constant."

SimpleConstantExpressionQ::usage = "Tests for Simple Constant."


Begin["`Private`"]


SetAttributes[Analyze, { HoldFirst }]

Analyze[expr_] :=
	Module[{ },
		Which[
			QuantumFreeQ[Evaluate[expr]],
			Print["No Quantum object detected."],
			BraExpressionQ[expr],
			Print["Detected a valid Bra."],
			KetExpressionQ[expr],
			Print["Detected a valid Ket."],
			OperatorExpressionQ[expr],
			Print["Detected a valid Operator."],
			QuantumConstantExpressionQ[expr],
			Print["Detected a valid Quantum Constant."],
			True,
			Print["Quantum object misformed."]
		];
		True]

QuantumFreeQ[expr_] := 
	FreeQ[expr, QuantumAlgebra`Operator] && 
	FreeQ[expr, QuantumAlgebra`Bra] && 
	FreeQ[expr, QuantumAlgebra`Ket] &&
	FreeQ[expr, QuantumAlgebra`BraKet]

QuantumExpressionQ[expr_] := 
	QuantumConstantExpressionQ[expr] || 
	BraExpressionQ[expr] || 
	KetExpressionQ[expr] || 
	OperatorExpressionQ[expr]


(* ::Section:: *)
(* **************** Region Check for Bra Expression ************************************* *)

SetAttributes[BraExpressionQW, { HoldFirst }]

BraExpressionQW[expr: QuantumAlgebra`Bra[__]] := 
	Length[expr] <= 10

BraExpressionQW[Plus[__?BraExpressionQ]] :=
	True

BraExpressionQW[Sum[_?BraExpressionQ, _]] :=
	True

BraExpressionQW[
	NonCommutativeMultiply[
		 _?BraExpressionQ, 
		 _?OperatorExpressionQ 
	]] :=
	True

BraExpressionQW[Times[ _?BraExpressionQ, _?ConstantExpressionQ ]] :=
	True

BraExpressionQW[Times[ _?ConstantExpressionQ, _?BraExpressionQ ]] :=
	True

BraExpressionQW[QuantumAlgebra`Hermitian[_?KetExpressionQ]] :=
	True

BraExpressionQW[expr_Symbol?ValueQ] :=
	BraExpressionQ[Evaluate[expr]]

BraExpressionQW[_] := False


(* ::Section:: *)
(* **************** Region Check for Ket Expression ************************************* *)

SetAttributes[KetExpressionQW, { HoldFirst }]

KetExpressionQW[expr: QuantumAlgebra`Ket[__]] := 
	Length[expr] <= 10

KetExpressionQW[Plus[__?KetExpressionQ]] :=
	True

KetExpressionQW[Sum[_?KetExpressionQ, _]] :=
	True

KetExpressionQW[
	NonCommutativeMultiply[ 
		_?OperatorExpressionQ, 
		_?KetExpressionQ 
	]] :=
	True

KetExpressionQW[Times[ _?KetExpressionQ, _?ConstantExpressionQ ]] :=
	True

KetExpressionQW[Times[ _?ConstantExpressionQ, _?KetExpressionQ ]] :=
	True

KetExpressionQW[QuantumAlgebra`Hermitian[_?BraExpressionQ]] :=
	True

KetExpressionQW[expr_Symbol?ValueQ] :=
	KetExpressionQ[Evaluate[expr]]

KetExpressionQW[_] := False


(* ::Section:: *)
(* **************** Region Check for Operator Expression ******************************** *)

SetAttributes[OperatorExpressionQW, { HoldFirst }]

OperatorExpressionQW[expr: QuantumAlgebra`Operator[__]] := 
	Length[expr] <= 4

OperatorExpressionQW[expr: (head: QuantumAlgebra`Operator[__])[__]] :=
	OperatorExpressionQ[head]

OperatorExpressionQW[Plus[__?OperatorExpressionQ]] :=
	True

OperatorExpressionQW[Sum[_?OperatorExpressionQ, _]] :=
	True

OperatorExpressionQW[
	NonCommutativeMultiply[ 
		_?OperatorExpressionQ, 
		_?OperatorExpressionQ 
	]] :=
	True

OperatorExpressionQW[
	NonCommutativeMultiply[ 
		_?KetExpressionQ, 
		_?BraExpressionQ 
	]] :=
	True

OperatorExpressionQW[Times[ _?OperatorExpressionQ, _?ConstantExpressionQ ]] :=
	True

OperatorExpressionQW[Times[ _?ConstantExpressionQ, _?OperatorExpressionQ ]] :=
	True

OperatorExpressionQW[Power[ _?OperatorExpressionQ, _Integer?NonNegative ]] :=
	True

OperatorExpressionQW[Power[ E, _?OperatorExpressionQ ]] :=
	True

OperatorExpressionQW[head_[_?OperatorExpressionQ]] /;
	ContainsAny[Attributes[head], { NumericFunction }] :=
	True

OperatorExpressionQW[QuantumAlgebra`Hermitian[_?OperatorExpressionQ]] :=
	True

OperatorExpressionQW[_?CommutatorExpressionQ] :=
	True

OperatorExpressionQW[expr_Symbol?ValueQ] :=
	OperatorExpressionQ[Evaluate[expr]]

OperatorExpressionQW[_] := False


(* ::Subsection:: *)
(* **************** Region Check for Commutators **************************************** *)

SetAttributes[CommutatorExpressionQW, { HoldFirst }]

CommutatorExpressionQW[
	QuantumAlgebra`Commutator[
		_?OperatorExpressionQ, 
		_?OperatorExpressionQ
	]] :=
	True

CommutatorExpressionQW[Times[ _?SimpleConstantExpressionQ, _?CommutatorExpressionQ ]] :=
	True

CommutatorExpressionQW[expr_Symbol?ValueQ] :=
	CommutatorExpressionQ[Evaluate[expr]]

CommutatorExpressionQW[_] := False


(* ::Section:: *)
(* **************** Region Check for arbitrary Constants ******************************** *)

SetAttributes[ConstantExpressionQW, { HoldFirst }]
SetAttributes[SimpleConstantExpressionQW, { HoldFirst }]
SetAttributes[QuantumConstantExpressionQW, { HoldFirst }]


ConstantExpressionQW[expr_] := 
	SimpleConstantExpressionQ[expr] || 
	QuantumConstantExpressionQ[expr]


(* ::Subsection:: *)
(* **************** Region Check for Simple Constants *********************************** *)


(* ::Subsubsection:: *)
(* Region Simple Constants 1 : what is SimpleConstant not *)

(*
SimpleConstantExpressionQW[expr_] :=
	FreeQ[expr, QuantumAlgebra`Bra] &&
	FreeQ[expr, QuantumAlgebra`Operator] &&
	FreeQ[expr, QuantumAlgebra`Ket]
*)

(* ::Subsubsection:: *)
(* Region Simple Constants 2 : what is SimpleConstant (positive test) *)


SimpleConstantExpressionQW[expr_] /;
	!FreeQ[expr, QuantumAlgebra`Bra] ||
	!FreeQ[expr, QuantumAlgebra`Operator] ||
	!FreeQ[expr, QuantumAlgebra`Ket] :=
	False

SimpleConstantExpressionQW[expr_?NumericQ] := True

SimpleConstantExpressionQW[Subscript[expr_Symbol?(!ValueQ[#]&),__]] := True

SimpleConstantExpressionQW[_Symbol[expr__]] := 									(* 2017-02-21: added to allow for c[i] *)
	And@@Map[SimpleConstantExpressionQ, {expr}]

SimpleConstantExpressionQW[expr_Symbol?(!ValueQ[#]&)] := True

SimpleConstantExpressionQW[expr_Symbol?ValueQ] := 
	SimpleConstantExpressionQ[Evaluate[expr]]

SimpleConstantExpressionQW[Sum[_?SimpleConstantExpressionQ, _]] :=
	True

(*	improved, all unit tests passed
*)
SimpleConstantExpressionQW[expr_] /;
	!FreeQ[expr, _Symbol] :=
	Module[
		{ 
			rng,
			symbols, csymbols, rules, 
			fsymbols, fcsymbols, frules, funcConstantQ, 
			ncsymbols 
		},
		funcConstantQ[ex_] :=
			Module[{head},
				head = Head[ex];
				ContainsNone[{head}, { Sum, List, Range, Infinity }] &&
				ContainsNone[Attributes[Evaluate[head]], {NumericFunction}] &&
 				Length[DownValues[head]] == 0
			];
		
		symbols = Union[Cases[expr, _Symbol, Infinity]];
		fsymbols = Union[Cases[expr, _Symbol[__], Infinity]];
		csymbols = Select[symbols, SimpleConstantExpressionQ];
		fcsymbols = Select[fsymbols, funcConstantQ];
		ncsymbols = Complement[symbols, csymbols];
		rng = Range[Length[csymbols]];
		rules = Inner[#1 -> #2&, csymbols, rng, List];
		rng = Range[Length[fcsymbols]];
		frules = Inner[#1 -> #2&, fcsymbols, rng, List];
		DiagnosticPrint[
			"CSymbols : ", csymbols, 
			"\nRules : ", rules, 
			"\nNCSymbols : ", ncsymbols,
			"\nFCSymbols : ", fcsymbols,
			"\nFRules : ", frules 
		];
		Length[ncsymbols] == 0 && NumericQ[expr //.frules //. rules]
	]

SimpleConstantExpressionQW[expr_] := False
		


(* ::Subsection:: *)
(* **************** Region Check for Quantum Constants ********************************** *)


QuantumConstantExpressionQW[QuantumAlgebra`BraKet[bra: {__}, ket: {__}]] := 	(* 2017-02-21: changed from SimpleConstantExpression *)
	Length[bra] <= 10 && Length[ket] <= 10

QuantumConstantExpressionQW[
	NonCommutativeMultiply[ 
		_?BraExpressionQ, 
		_?KetExpressionQ
	]] :=
	True

QuantumConstantExpressionQW[
	NonCommutativeMultiply[ 
		_?BraExpressionQ, 
		__?OperatorExpressionQ, 
		_?KetExpressionQ
	]] :=
	True

QuantumConstantExpressionQW[Plus[ _?QuantumConstantExpressionQ, _?ConstantExpressionQ ]] :=
	True

QuantumConstantExpressionQW[Plus[ _?ConstantExpressionQ, _?QuantumConstantExpressionQ ]] :=
	True

QuantumConstantExpressionQW[Sum[ _?QuantumConstantExpressionQ, _ ]] :=
	True

QuantumConstantExpressionQW[Times[ _?QuantumConstantExpressionQ, _?ConstantExpressionQ ]] :=
	True

QuantumConstantExpressionQW[Times[ _?ConstantExpressionQ, _?QuantumConstantExpressionQ ]] :=
	True

QuantumConstantExpressionQW[expr_Symbol?ValueQ] :=
	QuantumConstantExpressionQ[Evaluate[expr]]

QuantumConstantExpressionQW[_] := False


(* ::Section:: *)
(* **************** Region Diagnostic Print ********************************************* *)

(*	Level Handling for Recursive Predicates
	-> Updates level
	-> performs a diagnostic printout
*)
SetAttributes[BraExpressionQ, { HoldFirst }]
SetAttributes[KetExpressionQ, { HoldFirst }]
SetAttributes[OperatorExpressionQ, { HoldFirst }]
SetAttributes[CommutatorExpressionQ, { HoldFirst }]
SetAttributes[ConstantExpressionQ, { HoldFirst }]
SetAttributes[QuantumConstantExpressionQ, { HoldFirst }]
SetAttributes[SimpleConstantExpressionQ, { HoldFirst }]


BraExpressionQ := LevelUpdater[BraExpressionQW]

KetExpressionQ := LevelUpdater[KetExpressionQW]

OperatorExpressionQ := LevelUpdater[OperatorExpressionQW]

CommutatorExpressionQ := LevelUpdater[CommutatorExpressionQW]

ConstantExpressionQ := LevelUpdater[ConstantExpressionQW]

QuantumConstantExpressionQ := LevelUpdater[QuantumConstantExpressionQW]

SimpleConstantExpressionQ := LevelUpdater[SimpleConstantExpressionQW]


SetAttributes[LevelUpdater, { HoldFirst }]
SetAttributes[DiagnosticPrint, { HoldFirst }]


LevelUpdater[pred_] := 
	(Module[{result},
		level++;
		result = pred[Unevaluated[#]];
		DiagnosticPrint[#, pred, result];
		level--;
		result
	]&)

DiagnosticPrint[expr_, pred_, result_] :=
	Module[{},
		DiagnosticPrint[
			"Checked ", StringDrop[SymbolName[pred], -1], 
			"[ ", HoldForm[expr // FullForm], " ]", 
			" at level ", level, 
			": ", result
		];
		If[level == 1, QuantumAlgebra`Options`Private`QAPrint[1]]
	]

DiagnosticPrint[expr___] :=
	Module[{indent},
		indent[] := StringJoin[Table["  ", level - 1]];
		QuantumAlgebra`Options`Private`QAPrint[1, indent[], expr];
	]


level = 0;

End[]

EndPackage[]
