(* Wolfram Language Test file *)

(* Test of the Commutator Rule, Normal Ordering *)
(* We use the following parameters here:
   
   dim - the dimension
   up - flag, if we use upper indices
   normal - flag, if we produce normal ordering
   continuous = flag, if it is the continuous case
   on1, on2 - the operator names, they decide if the operators commute
   
   With the code generation approach we can vary those parameters, the rest remains the same.
*)
Test[
	(ToExpression["dim = 2"];																(* parameters *)
	up = True;
	normal = True;
	continuous = True;
	anti = False;
	on1 = p;
	on2 = p;
	ui1 = { 1 };
	ui2 = { 1 };

	delta = If[continuous, DiracDelta, DiscreteDelta];										(* preparation *)
	sgnanti = If[anti, -1, +1];
	sgndelta = Piecewise[{{ +1, anti }, { +1, normal }}, -1]; 								(* If[anti, +1, If[normal, +1, -1]]; *)
	com = If[
		Order[on1, on2] == 0 && Order[ui1, ui2] == 0, 
		delta@@Table[QASimplescript[l, i, up] - QASimplescript[k, i, up], {i, 1, dim}] Operator[1], 
		0];
	op1 = Operator[on1, Table[QASimplescript[k, i, up], {i, 1, dim}], ui1];
	op2 = Operator[on2, "H", Table[QASimplescript[l, i, up], {i, 1, dim}], ui2];
	{ op1, op2 } = If[normal, { op1, op2 }, { op2, op1 }];
	
	(op1 ** op2) /. DefineCommutatorRule[anti, continuous, normal, dim, up]),				(* final calculation *)
	
	sgnanti (op2 ** op1) + sgndelta com,													(* intended outcome *)
	
	TestID->"Exchange-20181226-W7C9S6"
]

(* Test of the Commutator Rule, Reverse Ordering *)
Test[
	(op1 = Operator[p, { QASimplescript[k, 1, True], QASimplescript[k, 2, True] }, {}];
	op2 = Operator[p, "H", { QASimplescript[l, 1, True], QASimplescript[l, 2, True] }, {}];
	(op2 ** op1) /. DefineCommutatorRule[False, False, False, 2]),
	(op1 ** op2) - DiscreteDelta[QASimplescript[l, 1, True] - QASimplescript[k, 1, True], QASimplescript[l, 2, True] - QASimplescript[k, 2, True]] Operator[1],
	TestID->"Exchange-20181226-K2G1Y6"
]

(* Test of the Anti Commutator Rule, Normal Ordering *)
Test[
	(op1 = Operator[p, { QASimplescript[k, 1, True], QASimplescript[k, 2, True] }, {}];
	op2 = Operator[p, "H", { QASimplescript[l, 1, True], QASimplescript[l, 2, True] }, {}];
	(op1 ** op2) /. DefineCommutatorRule[True, False, True, 2]),
	- (op2 ** op1) + DiscreteDelta[QASimplescript[l, 1, True] - QASimplescript[k, 1, True], QASimplescript[l, 2, True] - QASimplescript[k, 2, True]] Operator[1],
	TestID->"Exchange-20181226-J1P9F2"
]

(* Test of the Anti Commutator Rule, Reverse Ordering *)
Test[
	(op1 = Operator[p, { QASimplescript[k, 1, True], QASimplescript[k, 2, True] }, { 1 }];
	op2 = Operator[p, "H", { QASimplescript[l, 1, True], QASimplescript[l, 2, True] }, { 1 }];
	(op2 ** op1) /. DefineCommutatorRule[True, False, False, 2]),
	- (op1 ** op2) + DiscreteDelta[QASimplescript[l, 1, True] - QASimplescript[k, 1, True], QASimplescript[l, 2, True] - QASimplescript[k, 2, True]] Operator[1],
	TestID->"Exchange-20181226-C5C3G1"
]

(* Test of the Commutator Rule, Normal Ordering *)
(* The same as above, but with different operator names *)
Test[
	(op1 = Operator[p, { QASimplescript[k, 1, True], QASimplescript[k, 2, True] }, {}];
	op2 = Operator[q, "H", { QASimplescript[l, 1, True], QASimplescript[l, 2, True] }, {}];
	(op1 ** op2) /. DefineCommutatorRule[False, False, True, 2]),
	(op2 ** op1),
	TestID->"Exchange-20181226-B9M6B0"
]

(* Test of the Commutator Rule, Reverse Ordering *)
(* The same as above, but with different operator names *)
Test[
	(op1 = Operator[p, { QASimplescript[k, 1, True], QASimplescript[k, 2, True] }, {}];
	op2 = Operator[q, "H", { QASimplescript[l, 1, True], QASimplescript[l, 2, True] }, {}];
	(op2 ** op1) /. DefineCommutatorRule[False, False, False, 2]),
	(op1 ** op2),
	TestID->"Exchange-20181226-C0G6I7"
]

(* Test of the Anti Commutator Rule, Normal Ordering *)
(* The same as above, but with different operator names *)
Test[
	(op1 = Operator[p, { QASimplescript[k, 1, True], QASimplescript[k, 2, True] }, {}];
	op2 = Operator[q, "H", { QASimplescript[l, 1, True], QASimplescript[l, 2, True] }, {}];
	(op1 ** op2) /. DefineCommutatorRule[True, False, True, 2]),
	- (op2 ** op1),
	TestID->"Exchange-20181226-U2L6C8"
]

(* Test of the Anti Commutator Rule, Reverse Ordering *)
(* The same as above, but with different upper indices *)
Test[
	(op1 = Operator[p, { QASimplescript[k, 1, True], QASimplescript[k, 2, True] }, { 1 }];
	op2 = Operator[p, "H", { QASimplescript[l, 1, True], QASimplescript[l, 2, True] }, { 2 }];
	(op2 ** op1) /. DefineCommutatorRule[True, False, False, 2]),
	- (op1 ** op2),
	TestID->"Exchange-20181226-U0M3W9"
]
