(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 11.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[     93907,       2309]
NotebookOptionsPosition[     66150,       1634]
NotebookOutlinePosition[     90825,       2206]
CellTagsIndexPosition[     90782,       2203]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Maxwell Equation", "Title",
 CellChangeTimes->{{3.753095286161633*^9, 3.753095287559928*^9}, {
  3.7653000603160677`*^9, 3.765300065840293*^9}}],

Cell[CellGroupData[{

Cell["Initialization", "Section",
 CellChangeTimes->{{3.7530953033017855`*^9, 3.7530953157046394`*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"<<", "QuantumAlgebra`"}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.7464441987103815`*^9, 3.746444236622961*^9}, {
  3.7467845915842257`*^9, 3.7467846033962855`*^9}, {3.746784661775163*^9, 
  3.746784664075014*^9}}],

Cell[CellGroupData[{

Cell[BoxData["\<\"Application Quantum Algebra by Juergen Habelt\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJwd03s0FFgcB/Apj/KaGLQtKknSVIyQTA0mUlEWobJWpWwZz0ShkvJMHoWK
UnquymCTQhnl0Fsp5lRqUoMtMSFK65H2+9s/7vmc3/397j333vO7M/xC3f3H
MRiMYYwegfbSXYPdfGlsrbQAMiwcq87DKUkGatu/d/PXRQgCw+Gl/NNNLGUZ
/4KoXKwL8x+3RuhBncTiTRbwaIXcfku4zir5Ghn0c4KGNQw9p9zDhW3mfoO2
NF8V72UHY+JnvSVdv11TWwrr708zPwIdqoufZsHMRreF2XC5Y2/vUbj49Ojw
CcjpzVS7ClNiYySLVWR8+eKNPUtgf9h1NyFM9dI3LIFlr4YPlEK2VLH8Bmze
GS1fBfUWK+iQYwbP3pFKzIjo23BQL0/hDnSaGMwlU9gRh8liq/DoTmhX9iVL
UVXGf1nwypkJP1ppl6vDkCr5tnlQJoh1s4SiRbX9C2Gf4Zz3VtDncuZbLrxQ
+rJ+CWxLuJXDo3zT3jNk6opqh00wzm7RZvJkTEPPZahof25rBTR8mOzxDp4v
kIiV1GT8aceMbJShR2Sdtxps2Jt91Bj6GXUosKF6osV98vcQi3RrKM+eL+BC
wxRhMg9y+KEfyE8zk4fdYKqPd587tEofmuoJNyrmzSWv3GvIWQf3+L/pJG2t
7tmsh/s6Ff19oMnFgSxfWMvKGCBPzjjC2gBzEvfUF8LYdkkLec/Wzf0SfGpk
OOUynKjrY1AEY5Rzw0iHXjklIVwRv9+MNKzRXEpe93peRXLN216QWWuGXYph
kIbeblLJ+fIouWbc0SklUBhgt4HUibsbS4YXhheSah6ut0ibDdIPJOM916MU
Vkt4AaSxhBVN7jpUM/I3PGE6c/dVKLq/s6UMqvKu/CRdxs8LvUb3qq4Vk+qz
xpZU0zvIejQ74F32y23k2byuC2TW7fG//gN9c+64krdN/iwhFRlySSNw4M2q
WlKk4dVFGll6ThqFSfMqnEi3I4b1Kkz0deb0i6qQoaqSrgGTuGbl5IrmShMj
eM7Sb4Asi+XLZsNXH3MEbLjKyXrTXOhhmZBlCj80bqjkwCVa1lvMKH+8ieEF
WbM3N8xnyfjdv00/k62F2P5UVw5MC5+5+zjsj/+hn0uK5FQKoDvjs6WI8tKz
o1/h69oI7jc4P6aYPQiNR07+8R36PYqMJD3bWhf8C9XfGFeTKfP1G1naqIvk
mWhCSYHjZB4ZYBW5HL7yXLl2BcV7WqpIrXZ+lBMMH68yQOqLhL84w4zB5Vbk
aAg/aRVULbyUFEL7Ki3rJ7lPGvNDKf528wtZtJ3PlzrI+Ez9bMYnyLAP43fB
Hl7NjG74oOOSfC+0m7NnsB/WPQyepLgM/3aIU6gB47Sj9vHgwe1ReeTXBocF
3jTfOeRClgVG15Ktk0t/+sBnU2wH0uHrHaFppdAgVdZYDsVrj9VVQq5l7Iyb
8K+h4qFbsGXUzqAaPgjb8Ygcn7qmj4yOO58gghk1rGPkuaLwx+RY+v1vZKtj
GasGng/3jSAdvLVs62Fl10gfKShM4LRAw02TZ7bBYaNIJzLkulhA3nbpVGiH
RRkDJh+h/baM4E4YbD+hrovWc9z7yDrJqdJuaHNXSUqyBSOre+DNQKfrZFv6
I2YvdD2kl0k2Ozcz+6hOVO5FLnuxvqjv//PbSEnPilumco4y/px2ozx5KJR4
/yDHVvLYE6A4uGmaEnQY851gCm3qfGeRlYnv/Ukfc5PnHNjWJNA1g2G7py8g
G4T3Vi+AzHn9xeSYAA0IX4qvbjGHJUGaIlJYupm1nvZVNuj3hacNSiUBcFFQ
b4AAeufK3wikeGv+5ziof6Uk4ABM8YtSiocuvFydJNjKZXNIzlTPk2RkfYVK
MnTubCgnK5ZKtQ9D08/RbNJm+sOFWTAhNMKNFL+NrcyGTU+Yk3NhtE6aLpmW
pDtAMttTtPNgx8EdyeTqxrnxJ6BylVYPmShee6oIlu3P9RHC/wAiqblq
  "]],

Cell[BoxData["\<\"This package is based on the work of Cesar Guerra \
G.,\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJwd03s0FFgcB/AJkdeUCduiSFJNxZSkpgYTqSiLUFmrUm2Z8kgUkpRnCuVR
RClqkcEmhRjl0FspzanUpIa2xISI1iPt97d/3PM5v/v73d+59557p3v7u+wY
x2AwhjG6BVorDgx28aURddIcyFhkV5UHp8Qaqu/93sXfGCTYHQgLss83s1Rk
/EuicrEuzH7UGqQHdWKKty6C6RXyR8zhRou4a+Sen0oaS6F/rko3F7aZeQ9a
0XxVlLs1DIua+ZZ0GrimvgI23Jtmdgra1hQ/SYHJTc6LU+Equ56edLjs/Ojw
WcjpSVa/CuMjwiTLVGV8heIt3cthX8B1ZyFMcDcwKoFlr4aPlkK2VLH8Bny+
P1ShCuotG69Djhk+fUcqM4NCb8FBvczxt6H9BF8uGc8OOkkWWwSGdkDrsq8p
imoy/sucVw5M+MlCq3wS9KtSaJsHZYIIZ3MoWlLXtxj2Gs15bwE9C5PfcuGl
0pcNy2FbdHUaj/LNhy6QCatrbLfCSOsl28issMbuQqhok7uzAho9iHN9B/Ny
JGJldRl/2mljSxXoGlzvoQ4bD6Wmz4bexh/Gs+GkmEX3yN/9FiUuhQrs+QIu
NIoXxvEgh+//kfw8I27YGSZ4evS6QIvEoalucIti5lzyyt3GtI0wfMebDtLK
4q7lJni4Q3GHJzS53J/iBetYSf1k1vRTrM0wLSa8IR9GtEtayLtWzi4F8Imx
0ZRCOEHX07AIhqlkBJC2PfLKQrg66sgC0qh28gryuvuzKpJr1vaCTFk/7FgM
92joHSSVHQpHyfXj0qeUQKGP9WZSJ/JOBBmYH5hPqrs6VZOWm6UfScZ7rmsp
rJHwfMjZElYoeeB47cjf8KzpjINXoeje/pYyqMa78pN0lJvnf43OVVMnJifN
HFteQ/cg6578Ad5hv9xFXszsvESm3JL79R/olXbbibxl8mcJqciQjx2B/W/W
1pEiDfdO0tjcbeIojJ1XYU86nzJqUGXiXSfrX1aDDDXVRA0Yy11QTq5+Xmli
DHPNvfvJsgi+bBZ89SlNwIZr7ZdunQtdzaNTTOHHps2VHLhcc+n2BZQ/08xw
h6xZ2xrns2T8rt/0L6RqIrY515kGTwTOOHgG9kX9MMggRfKqOdCF8cVcRHnp
xdFv8HVdEHcAzg8rZg/C2SNZf3yH3g+Dg0m3ttaF/8JJb2bXkPHzDZpYWqgL
5plMhpIcO20e6WMRvAq+cluzYTXF4S1VpGY7P8QeBsqp9pMGIuEvDjBpcJUF
OerHj10L1fILYv2or/LKPpL7uCnbn+KBm1/Jor18vtRWxmcapDI+Q4ZNAL8T
dvNqp3fB+x8KFHqg9ZzwwT5Y/8B3ouJK/NshTr4GjNQKOcyDx/aGZJLfGm0X
etB8x5AjWbY7tI5s1S796QmfTrHqT4Sv9/mfKIWGCbKmcijecLq+EnLNI6bf
hH8NFQ9Vw5ZRa8MaeD9g30NSLmF9LxkamRctgkm1rNNkblHgI3Is8d4A2WpX
xqqFeYFeQaSth6ZVA6zsHOklBfnRnBZotFV7RhscNg62J/2uiwXkLceO8e2w
KKnf5BO02ZXk2wF9bZTqO2k9x6WXrJecK+2ClneUpSRbMLKuG97cbX+dbEt8
yOyBTsf1ksnnDs+ZvVQnKncnV77YVNT7//4tpaRbRbWpvJ2MP6fdOFMBCiUe
P8ixNTy2EhT7Nk9ThrZjXkqm0LLeayZZGfN+B+lpZvKMA9uaBboLYMBB/YVk
o/DuuoWQOa+vmBwT4AHCl+Kr281gyZ7JIlJYuo21ifqqGPZ5wfOGpRIfuGRP
j48AemQo3NhN8c7sL5HQ4EqJz1EY7x2iHAUdeRk6sbCVy+aQnKluWWRwQ4Vq
HHToaCwnK1ZItU5C0y+hbNJS/8HiFBjtH+RMit9GVKbC5sdM7QwYqnNClzwR
q9tPMtvjtTLhh2P74sh1TXOjzkKVKs1uMka84VwR1V/J8BTC/wAimLlt
  "]],

Cell[BoxData["\<\"who developed the original package for Mathematica \
4.1/4.2\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJwd03s0FFgcB/ApUl4Tg7ZFJSRNYoTU1GAiibIIlbUqZcuUR6KikvJMHoWK
UopalcEmQhnl0Fsp5lRqUkNbYkKU1iPt97d/3PM5v/v73Xvuved3Z/qHeASM
YzAYwxg9Aq2luwe7+dLoOmkeZFg6VhfAqQn6qju+d/PXhgu2hcFLuWebWUoy
/gVRuVgH5j5qC9eF2vHFGy3h8Uq5g1ZwrXXiNXL7z4nqi2BIvlIPF7Zb+A/a
0nx1rLcdjIqd9YZ0+3ZNdSlsuDfd4hh0qCl+kgHTm9wXZMLljr29x+His6PD
pyCnN131KkyKjpIsVpbx5Ys39CyB/aEV7kKY7K1nWALLXg4fKoVsqUL5ddiy
K1K+GuounqBNjuk/fUsqMsMjb8FB3ZwJt6HzpCAumcQOP0oWW4dFdkK7si8Z
Cioy/ou8ly5M+NFaq1wNBlfLt5tAmSDa3QqKFtb1L4B9hnPeWUPfy+lvuPBC
6YuGJbA97mYWj/LN+8+RyU41DhthjN3CTeTpqMaey1DBPn9LJTR8kOj5Fhbk
ScSKqjL+9BNGNkrQM6LeRxU27s88bgz9jd5PYEO1eMt75O/BlqmLoDx7noAL
DZOEiTzI4Yd8ID8ZJA67w2Rfnz4PaJ06NM0LblDImUteuduYtRbuC3jdSdpa
37VZBw90KgT4QtOLAxl+sI6VNkCennmMtR5mxe9rKITRHZJW8q6tu8cl+MTI
cOplOEnHV78IRillh5IOvXKKQugUe9CcNKzVWEpWeD+rJrkW7c/JjNXDrsVw
u7ruXlLR5fIouXrc8aklUBhot57UjrkTTYYVhhWSqp5uN0mb9dIPJOMd17MU
1kh4gaSxhBVJ7j5SO/I3PGVmsPcqFN3b1VoGVXhXfpKu401CrtG9aurEpNqs
sSU19A6yHo338A77xVbyfE7XBTLj1vhf/4F+WbfdyFumf5aQCgy5hBE48Hpl
HSlS9+4ijay8Jo/CBJNKZ9L9mGGDMhN9nT7jogpkqCinqsMErnk56dRSZWoE
8638B8iyaL5sNnz5MUvAhiudF22cCz2t4jLM4Iem9VUcuERz0WZzyp9sZnhD
1uxNjfNYMn73bzPOZWoitj/TlQVTwgz2noT9sT/0skmRnHIe9GB8thJRXnp+
9Ct8VRfO/QbnRRWzB6HxyOk/vkP/hxERpFd72/x/odpr4xoyaZ5eE0sLdRE8
Uw0oyXOcwiMDrSOWw5deK9Y4UbyvtZrU7ODvcYZh45UHSD2R8BcXmDa43Joc
DeYnrIQqhZcSgmlfxWX9JPdxU24Ixd9ufCGLdvD5UgcZn6mXyfgEGfah/C7Y
w6ud2Q3vv78k3wvt5uwb7If1D4ImKyzDvx3iFKrDGK09B3jw8I49OeTXRof5
PjTfOeRKlm2LrCPbppT+9IVPp9oOpMJXO0NSSqF+sqypHIrXnKivglyr6Jk3
4F9DxUM3YeuonX4NvB+68yE5Pnl1HxkZUxAngmm1rBNkflHYI3Is9d43ss2x
jFULC8L8wkkHH03bBljVNdJHCgrjOK3QcOMUg3Y4bBThTAZXiAXkLdfOCR2w
KG3A9CO035oW1AmD7CfWd9F6jkcfWS85U9oNbe4oSkm2YGRVD7yxzbmCbE99
yOyFbkd008kWlxZmH9WJyr3JZc/XFfX9f34bKelVedNMzlHGn9NhlCMPhRKf
H+TYCh57IhQHNU9XhA5jfhPNoE293yyyKv5dAOlrYfqMA9ubBTrmMHTvjPlk
o/DuqvmQadJfTI4J0IDwhfjqZgtYsl1DRApLN7HW0b5K+v1+8Kx+qSQQLtze
GyiAPtny17dRvCX3cwzUu1ISeAgm+e9RjIWuvGztBNjGZXNIzjSv02REQ6Vy
InTpbCwnK5dKtY5Cs8+RbNJmxoMFGTAuJNydFL+JrsqEzY+ZU7JhpHaKDpmS
oDNAMjuStHLg+8M7E8lVTXNjT0Glas0eMl685kwR7OzO9hXC/wAnP7oK
  "]],

Cell[BoxData["\<\"For proper styling you should invoke SetNotebookStyles \
once.\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJwd03tUDFgcB/DRi15Do1pbIUkY1JDEaKpRQtFWKrRtCEvTU4peEj1FRYXy
irBJU62USk10yjsicyhGTFlSo1LK9pD9/vaPez7nd3+/e8+99/zuDO9Alx3j
GAzGMEa3QGvFvsEuvjS6VpoDGYvtKi/BKQkG6ru/d/E3hgh8g+HVs+ebWCoy
/mVRqVgXnn3cGqIHdeILty6GJ8rlD5rBjeaJN0i/n+M1lsHAXJVuLmwz9R60
ovnKWHdrGBE76y3pNHBDfQWsvz/N9Di0rS58mg7TGp2XZMBVdj09J+Dy86PD
pyGnJ039OkyKjpAsV5XxFQq3dFvAvqAyZyFMdtc3LIIlzcOHiiFbqlR6E77Y
G65QCfWWK+qQYwbP3pHKzJDw23BQL1vxDrSf4M8lk9ghx8hC8+DwDmhd8jVd
SU3Gf5XT7MCEn8y1SifBgEqFtvlQJoh2NoOipbV9S2Cv4dz35tAzP+0tF14u
flVvAdviqjJ5lG/af4FMXl1tuxXGWC/dRp6JaOjOh0o2uTvLoeHDRNd38FKO
RKysLuNPO2lkqQJdQ+s81GHD/owTc6C30QdFNpwUv/g++XvA4pRlUIG9QMCF
hknCRB7k8AM/kp9nJg47w2RPj14XaJ4yNNUNblHKnkdeu9eQuRFG7XjTQVqZ
37PcBA90KO3whMZX+tO9YC0rtZ88M+M4azPMjI+qz4PR7ZIW8p6Vs8tV+NTI
cEo+nKDraVAAI1SygkjbHnllIVwde3AhaVgzeQVZ5v68kuSatr0k09cPOxZC
Pw29SFLZIX+UXD/uxJQiKPSx3kzqxNyNJoPzgvNIdVenKtJys/QjyXjPdS2G
1RKeDzlHwgon9x2pGfkbnjaZGXkdiu7vbSmBarxrP0lHufmBN+he1bVictKs
MYtqegdZ9+QP8C771S7yYnbnZTL9ttyv/0CvzDtO5G3jP4tIJYZ8wgjsf7O2
lhRpuHeSRmZuE0dhwvxye9L5uGG9KhN9nTb9ihpkqKmmaMAE7sJScvWLCmMj
mGvm3U+WRPNls2Hzp0wBG661X7Z1HnQ1i0s3gR8bN1dwoIXmsu0LKX+qieEO
WbO3NSxgyfhdv02/kKGJ2OZcZyY8Gjwz8hTsi/2hn0WK5FVzoAvji5mI8tKL
o9/g69oQ7gBcEFHIHoRzRs788R16PwoNJd3aWhf9Cye9mVNNJi3Qb2RpoS6U
ZzwZSnLstHmkj3noKtjstmbDaoqjWipJzXZ+mD0MllPtJ/VFwl8cYOrgKnNy
NICfsBaq5V1NCKB9lVf2kdwnjWcDKR649ZUs2M3nS21lfKZ+BuMzZNgE8Tth
N69mRhd88OGqQg+0nhs12AfrHvpPVFqJfzvEydOAMVphB3jw8O6wbPJbg+0i
D5rvGHIkS3zDa8lW7eKfnvDZFKv+FPh6T+DRYmiQLGssheINJ+sqINcsesYt
+NdQ4VAVbBm1NqiGD4L2PCLlktf3kuExl+JEMLWGdZLMLQh+TI6l3B8gW+1K
WDXwUrBXCGnroWlVDys6R3pJQV4cpwUabtWe2QaHjULtyYAysYC87dih2A4L
UvuNP0GbXan+HdDfZnxdJ63nuPSSdZJzxV3Q8q6ylGQLRtZ1w1u+9mVkW8oj
Zg90OqKXRr5weMHspTpRqTu58uWmgt7/z28pJd3Kq0zk7WT8ue1G2QpQKPH4
QY6t4bHHQ7F/0zRlaDvmNd4EWtZ5zSIr4t/vID1NjZ9zYFuTQHchDIqcvohs
EN5btwgy5/cVkmMCNCB8Jb6+3RQW+U0WkcLibaxNtK+KQZ8XPG9QLPGBS/16
fATQI0vhpi/FO89+iYH614p8DsEk7zDlWOjIy9JJgK1cNofkTHU7Q4bWl6sm
QoeOhlKyfIVU6xg0+RLOJi2nP1ySDuMCQ5xJ8dvoigzY9ISpnQXDdY7qkkcT
dPtJZnuSVjb8cHhPIrmucV7saahSqdlNxos3nCuAOrOyPYXwPxtruHc=
  "]],

Cell[BoxData["\<\"ALIASES:\\nESC op(du) ESC - Operator with or without \
indices\\nESC hop(du) ESC - Hermitian Operator with or without indices\\nESC \
ket(du) ESC - Ket with or without indices\\nESC (du)bra ESC - Bra with or \
without indices\\nESC (du)braket(du) ESC - BraKet with (left and / or right) \
subscript(s) and / or superscript(s).\\nLetters in parenthesis signify \
optional indices (u - up, d - down) and must be given in that \
order.\\nExample: use opd for an Operator with lower index.\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJwd03s0FFgcB/ApIq+JCduiCKmmYiTJ1GAiibIRKmuJsmXKI1FeSXkmj0JF
Lz1XZbBJoQw59FZKcyo1qaEtMSFK65H2+9s/7vmc3/397j333vO70wNC3QPH
MRiMYYwegdbSXYPdfGl8vbQQMhY4Vp+DU1IM1bZ/7+avixBsDYcXT5xqYSnL
+OdFFWJdeOJhW4Qe1Eku8V8AD1fK7bWE66xSr5LbfipqWMPQs8o9XNhuETBo
S/PViV52MCZxxhty9berakth491pFoegQ03J4xyY3ey2MBcud+ztPQwXnxod
PgY5vdlqV2BafIxksYqML1+yoWcJ7A+75iaE6V4GxqWw/OXwvjLIlipUXIfP
dkbLV0O9xRN0yDHDJ29JJWZEdB0c1CuYcAs6TwzmkmnsiINkiVV4dCe0K/+S
o6Aq478ofOnChB+ttCrUYUi1fPtcKBPEu1lC0aL6/oWwz3j2Oyvocyn7DRee
L3vRuAS2J93M41G+ZfdpMt2pxsEfJtgt2kgej2nquQQV7M9uroTG91M93sJz
hRKxkpqMP+2IiY0y9Ihs8FaDTbtzD8+CASbvJ7ChevKCu+TvIQsyraE8e56A
C43ThKk8yOGHfiA/GaUOu8F0H+8+d2iVOTTVE25QKJhDXr7TlLcOxgW+7iRt
re7YrId7OhUCfaDphYEcX1jPyhogj08/xPKDeclxjUUwvkPSSt6xdXO/CB+b
GE+5BCfq+hgWwxjl/DDSoVdOSQidEveak8a1k5eS17yeVpNci/bnZM6aYdcS
uE1DL5ZUcrk0Sq4Zd3hKKRQG2fmROgm348nwovAiUs1j9U3Sxk/6gWS843qU
wRoJL4icJWFFk7sO1I78DY+ZGcVegaK7O1vLoSrv8k/Sdfzc0Kt0r5p6Mak+
Y2xJDb2DrGfye3ib/WILeaag6zyZUzf+13+gb96t1WSd6Z+lpAJDLmUEDrxe
WU+KNLy6SBNLz0mjMGVupTPpdsi4UYWJvs7Wv6AKGaoqmRowhWteQTo9qzI1
gWctAwbI8ni+bCZ8+TFPwIYrna3950APy6QcM/ih2a+KA5doWm8yp/zRFoYX
ZM3c2DSPJeN3/6Z/OlcTsf3JrjyYEW4UexT2J/4wyCdFciqF0J3x2VJEeemZ
0a/wVX0E9xucF1PCHoSzRo7/8R0GPIiMJD3b2+b/C9Vfz6oh0+YZNLO0UBfJ
M50MJYWO2jwyyCpyOXzpuWKtE8VxrdWkZgc/yhmGj1cZIA1Ewl9cYNbgcity
NISfshKqFl1MCaF9lZb1k9xHzSdCKf524wtZvJ3PlzrI+EyDXMYnyLAP43fB
Hl7t9G547/1F+V5oNztusB823A+epLAM/3aIU6QBE7Si9vDg/u1RBeTXJof5
3jTfOeRKlm+NrifbtMt++sAnU2wHMuGrHaEZZdAwXdZcAcVrjzRUQa5l/PQb
8K+hkqGbsHXUzrAG3gvb8YAcn76mj4xOOJckglm1rCPk2eLwh+RY5t1vZJtj
OasWngv3jSAdvDVtG2FV10gfKShK4rRCY39to3Y4bBLpTIZcEwvIOtfOCR2w
OGvA9CO035IV3AmD7RUbumg9x72PbJCcLOuGNreVpCRbMLKqB97Y6nyNbM98
wOyFqw/oZZPPXJ4x+6hOVOFFLnu+vrjv//PbSEnPyptmco4y/uwOkwJ5KJR4
/yDHVvDYilAc3DJNCTqM+SqaQZsG3xlkVfK7QNLHwvQpB7a3CHTNYVis/nyy
SXhn1XzInNtfQo4J0IDwhfjKJgtYum2yiBSWbWStp32VDft94SnDMkkQXLSt
N0gAvfPlr2+lePOJzwnQ4HJp0D6YFhCllAhdefk6KbCNy+aQnKmex8nIxkqV
VOjS2VRBVi6Vah2EZp+j2aSN/v2FOTApNMKNFL+Jr8qFLY+Y2vkwWidDl8xI
0R0gmR1pWgXw/f4dqeSq5jmJx6BytWYPmSxee7IY6rh0+gjhfx5PuOw=
  "]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Problems", "Section",
 CellChangeTimes->{{3.765084024850484*^9, 3.765084029889999*^9}}],

Cell[CellGroupData[{

Cell["\<\
Certain operations are opponents of each other -> problem if used \
automatically\
\>", "ItemNumbered",
 CellChangeTimes->{{3.765084087946722*^9, 3.7650841407633977`*^9}, {
  3.7650992612456226`*^9, 3.7650992858690834`*^9}, {3.765109003918614*^9, 
  3.7651090269144745`*^9}, {3.7950635810017667`*^9, 3.795063617631888*^9}}],

Cell["\<\
Notation with QASimplescript: Implementation solves power problem\
\>", "ItemNumbered",
 CellChangeTimes->{{3.765084087946722*^9, 3.7650843110189657`*^9}, {
  3.7650992913529243`*^9, 3.765099291353916*^9}, {3.7651090683706026`*^9, 
  3.7651090959837413`*^9}, {3.795011605503235*^9, 3.7950116460478563`*^9}}],

Cell["DistributionLaw: Implementation does to change exponents", \
"ItemNumbered",
 CellChangeTimes->{{3.765084087946722*^9, 3.7650843110189657`*^9}, {
  3.765099294117523*^9, 3.7650993257808547`*^9}, {3.7651091167790794`*^9, 
  3.7651091546040864`*^9}, {3.795011568662716*^9, 3.7950115940229445`*^9}}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Rules", "Section",
 CellChangeTimes->{{3.7530953316250696`*^9, 3.7530953324568453`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{"Metric", ",", " ", 
    RowBox[{
     RowBox[{
     "used", " ", "for", " ", "calculations", " ", "with", " ", "co"}], "-", 
     " ", 
     RowBox[{"and", " ", "contravariant", " ", "components"}]}]}], " ", 
   "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    SubscriptBox["g", 
     RowBox[{"\[Mu]_", ",", "\[Nu]_"}]], ":=", 
    RowBox[{"Piecewise", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"1", ",", 
          RowBox[{"\[Mu]", "\[Equal]", "\[Nu]", "\[Equal]", "0"}]}], "}"}], 
        ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"-", "1"}], ",", 
          RowBox[{"\[Mu]", "\[Equal]", "\[Nu]"}]}], "}"}]}], "}"}], ",", 
      "0"}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    SuperscriptBox["g", 
     RowBox[{"\[Mu]_", ",", "\[Nu]_"}]], ":=", 
    RowBox[{"Piecewise", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"1", ",", 
          RowBox[{"\[Mu]", "\[Equal]", "\[Nu]", "\[Equal]", "0"}]}], "}"}], 
        ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"-", "1"}], ",", 
          RowBox[{"\[Mu]", "\[Equal]", "\[Nu]"}]}], "}"}]}], "}"}], ",", 
      "0"}], "]"}]}]}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.7531806315641794`*^9, 3.753180813885485*^9}, {
  3.753180868392724*^9, 3.7531809237755814`*^9}, {3.753181032701586*^9, 
  3.753181079309929*^9}, {3.7642321886778216`*^9, 3.764232292932061*^9}, {
  3.7650943246584063`*^9, 3.765094350330728*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "Looks", " ", "as", " ", "follows", " ", "when", " ", "using", " ", "the", 
    " ", "module", " ", "implementation"}], " ", "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"QASelectDefault", "[", "\"\<Vectors\>\"", "]"}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"QASetIndexCharacter", "[", 
     RowBox[{"{", 
      RowBox[{"\"\<Mixed\>\"", ",", 
       RowBox[{"{", "}"}]}], "}"}], "]"}], ";"}], "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"QAResetExplicitIntegrationVariables", "[", "]"}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"QAAddExplicitIntegrationVariables", "[", 
     RowBox[{"{", 
      RowBox[{
       SubscriptBox["k", "0"], ",", " ", 
       SubscriptBox["k1", "0"]}], "}"}], "]"}], ";"}], "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"QAResetAssumptions", "[", "]"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"QAAddAssumptions", "[", 
     RowBox[{"Map", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"#", "\[Element]", "Reals"}], "&"}], ",", 
       RowBox[{"Table", "[", " ", 
        RowBox[{
         SubscriptBox["k", "i"], " ", ",", 
         RowBox[{"{", 
          RowBox[{"i", ",", "0", ",", "3"}], "}"}]}], "]"}]}], "]"}], "]"}], 
    ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"QAAddAssumptions", "[", 
     RowBox[{"Map", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"#", "\[Element]", "Reals"}], "&"}], ",", 
       RowBox[{"Table", "[", " ", 
        RowBox[{
         SubscriptBox["k1", "i"], ",", 
         RowBox[{"{", 
          RowBox[{"i", ",", "0", ",", "3"}], "}"}]}], "]"}]}], "]"}], "]"}], 
    ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"QAAddAssumptions", "[", 
     RowBox[{"Map", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"#", "\[Element]", "Reals"}], "&"}], ",", 
       RowBox[{"Table", "[", " ", 
        RowBox[{
         SuperscriptBox["x", "i"], " ", ",", 
         RowBox[{"{", 
          RowBox[{"i", ",", "0", ",", "3"}], "}"}]}], "]"}]}], "]"}], "]"}], 
    ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"QAAddAssumptions", "[", 
     RowBox[{"Map", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"#", "\[Element]", "Reals"}], "&"}], ",", 
       RowBox[{"{", 
        RowBox[{"V", ",", "q"}], "}"}]}], "]"}], "]"}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"QAAddAssumptions", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        SubscriptBox["k", "0"], "\[GreaterEqual]", "0"}], ",", 
       RowBox[{
        SubscriptBox["k1", "0"], "\[GreaterEqual]", "0"}], ",", " ", 
       RowBox[{"V", "\[GreaterEqual]", "0"}]}], "}"}], "]"}], ";"}], 
   "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rule", "[", 
     RowBox[{"r_", ",", "s_"}], "]"}], ":=", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       SubsuperscriptBox["\[Epsilon]", 
        RowBox[{
         SubscriptBox["k_", "1"], ",", 
         SubscriptBox["k_", "2"], ",", 
         SubscriptBox["k_", "3"]}], 
        RowBox[{"0", ",", "r"}]], "*", 
       SubsuperscriptBox["\[Epsilon]", 
        RowBox[{
         SubscriptBox["k_", "1"], ",", 
         SubscriptBox["k_", "2"], ",", 
         SubscriptBox["k_", "3"]}], 
        RowBox[{"0", ",", "s"}]]}], ")"}], "\[RuleDelayed]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"r", "\[Equal]", "0"}], ",", 
           RowBox[{"-", "1"}], ",", 
           RowBox[{"+", "1"}]}], "]"}]}], "*", 
        RowBox[{"KroneckerDelta", "[", 
         RowBox[{"r", ",", "s"}], "]"}]}], "+", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"i", "=", "1"}], "3"], 
        RowBox[{"(", 
         RowBox[{
          SubsuperscriptBox["\[Epsilon]", 
           RowBox[{
            SubscriptBox["k", "1"], ",", 
            SubscriptBox["k", "2"], ",", 
            SubscriptBox["k", "3"]}], 
           RowBox[{"i", ",", "r"}]], "*", 
          SubsuperscriptBox["\[Epsilon]", 
           RowBox[{
            SubscriptBox["k", "1"], ",", 
            SubscriptBox["k", "2"], ",", 
            SubscriptBox["k", "3"]}], 
           RowBox[{"i", ",", "s"}]]}], ")"}]}]}], ")"}]}]}], 
   "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"QAAddReplacementRules", "[", 
     RowBox[{
      RowBox[{"Table", "[", 
       RowBox[{
        RowBox[{"rule", "[", 
         RowBox[{"i", ",", "j"}], "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"i", ",", "0", ",", "3"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"j", ",", "0", ",", "i"}], "}"}]}], "]"}], "//", "Flatten"}],
      "]"}], ";"}], "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"QAAddReplacementRules", "[", 
     RowBox[{
      RowBox[{"{", " ", 
       RowBox[{
        RowBox[{"DefineNameOrderingRule", "[", "False", "]"}], ",", 
        RowBox[{"DefineCommutatorRule", "[", 
         RowBox[{"False", ",", "True", ",", "True"}], "]"}]}], "}"}], ",", 
      "\"\<CommutatorLocal-Vectors\>\""}], "]"}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"QAAddReplacementRules", "[", 
     RowBox[{
      RowBox[{"{", " ", 
       RowBox[{
        RowBox[{"DefineNameOrderingRule", "[", "False", "]"}], ",", 
        RowBox[{"DefineCommutatorRule", "[", 
         RowBox[{"False", ",", "True", ",", "False"}], "]"}]}], "}"}], ",", 
      "\"\<CommutatorReverseLocal-Vectors\>\""}], "]"}], ";"}], 
   "\[IndentingNewLine]"}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.748970308237199*^9, 3.7489703878332844`*^9}, {
   3.7489704228220263`*^9, 3.7489704701583986`*^9}, 3.748970505174759*^9, 
   3.7489706999186296`*^9, {3.7489713455338335`*^9, 3.7489714433490677`*^9}, {
   3.7489715074835024`*^9, 3.7489715207165184`*^9}, {3.749012846909777*^9, 
   3.749012872213085*^9}, {3.7490384008539915`*^9, 3.749038475563145*^9}, {
   3.749038552595388*^9, 3.749038565637499*^9}, {3.7505949912854013`*^9, 
   3.7505950665186987`*^9}, {3.75060533266099*^9, 3.7506053386679497`*^9}, {
   3.750605374858466*^9, 3.750605379331471*^9}, {3.750605416585587*^9, 
   3.7506054252783766`*^9}, {3.750605602976101*^9, 3.7506056339863353`*^9}, {
   3.7507650100213957`*^9, 3.75076506169322*^9}, 3.75076619430978*^9, 
   3.75078655421354*^9, {3.753095519103619*^9, 3.753095520660427*^9}, {
   3.7530967603660183`*^9, 3.7530967613036656`*^9}, {3.7531881011524467`*^9, 
   3.753188154594534*^9}, {3.7531882606797504`*^9, 3.753188300619973*^9}, {
   3.75318838666881*^9, 3.7531885418730392`*^9}, {3.7531886609904575`*^9, 
   3.7531887033045917`*^9}, 3.7634622371455483`*^9, {3.7634627113551455`*^9, 
   3.7634627150981383`*^9}, {3.7634691901636467`*^9, 3.7634692118327427`*^9}, 
   3.763469257008788*^9, {3.763469307872674*^9, 3.763469351185894*^9}, {
   3.7634693859230337`*^9, 3.763469434597416*^9}, {3.7634694774338775`*^9, 
   3.7634694882130594`*^9}, {3.7634695292742805`*^9, 3.763469614839528*^9}, {
   3.7634698690413604`*^9, 3.7634698879946847`*^9}, {3.763470143659425*^9, 
   3.7634702109255543`*^9}, {3.7634709903575287`*^9, 
   3.7634709934313016`*^9}, {3.763471561819458*^9, 3.7634715837129235`*^9}, {
   3.7634717213093615`*^9, 3.763471724058016*^9}, {3.7634719604944644`*^9, 
   3.76347202417523*^9}, 3.763474251712321*^9, {3.76347433965926*^9, 
   3.7634743625710015`*^9}, {3.763474403518524*^9, 3.7634744141620646`*^9}, {
   3.7636352373924255`*^9, 3.76363524102275*^9}, {3.7636406614000688`*^9, 
   3.7636406938734326`*^9}, 3.763645541646773*^9, {3.7636462330927*^9, 
   3.7636462962422333`*^9}, {3.763701092027913*^9, 3.7637011277234783`*^9}, {
   3.7642314425544767`*^9, 3.764231596129636*^9}, {3.7642316741903806`*^9, 
   3.7642317133117294`*^9}, {3.7642317930364904`*^9, 
   3.7642318844219875`*^9}, {3.7642377661349087`*^9, 3.764237770885201*^9}, {
   3.764237816147149*^9, 3.7642378709996567`*^9}, {3.7642379028315997`*^9, 
   3.764238038626336*^9}, {3.764762947131156*^9, 3.764762994299947*^9}, {
   3.764763030242834*^9, 3.764763031440624*^9}, {3.7647630877040854`*^9, 
   3.7647631664613943`*^9}, {3.764763198655275*^9, 3.764763201393979*^9}, {
   3.7647633219803877`*^9, 3.764763341507127*^9}, 3.764763422963234*^9, 
   3.764846331134133*^9, {3.7688837153544416`*^9, 3.7688837437644377`*^9}, {
   3.768883803375267*^9, 3.768883911013529*^9}, {3.7688870538206377`*^9, 
   3.7688870736904755`*^9}, {3.768907634333249*^9, 3.7689077510290966`*^9}, {
   3.7692314505305705`*^9, 3.769231463381192*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"QAAddReplacementRules", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         SubscriptBox["k_", "1"], "^", "2"}], "+", 
        RowBox[{
         SubscriptBox["k_", "2"], "^", "2"}], "+", 
        RowBox[{
         SubscriptBox["k_", "3"], "^", "2"}]}], "\[RuleDelayed]", 
       RowBox[{
        SubscriptBox["k", "0"], "^", "2"}]}], ",", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         RowBox[{
          SubscriptBox["k_", "1"], "^", "2"}]}], "-", 
        RowBox[{
         SubscriptBox["k_", "2"], "^", "2"}], "-", 
        RowBox[{
         SubscriptBox["k_", "3"], "^", "2"}]}], "\[RuleDelayed]", 
       RowBox[{"-", 
        RowBox[{
         SubscriptBox["k", "0"], "^", "2"}]}]}]}], "}"}], "]"}], ";"}], 
  " "}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.7634609267462626`*^9, 3.763460934141495*^9}, {
   3.76346097380542*^9, 3.76346097737188*^9}, {3.7634610236142745`*^9, 
   3.763461081481556*^9}, 3.763461113619601*^9, {3.763470002824666*^9, 
   3.763470017683225*^9}, {3.7642320359656177`*^9, 3.764232064666813*^9}, {
   3.7688840616816053`*^9, 3.768884183248418*^9}, {3.769088712816906*^9, 
   3.769088724427804*^9}, {3.7690908983639374`*^9, 3.7690909308121405`*^9}, {
   3.7942308177977047`*^9, 3.794230817976231*^9}, {3.7942874777816153`*^9, 
   3.794287496854618*^9}, 3.79514799591633*^9}],

Cell[BoxData[
 RowBox[{"(*", " ", 
  RowBox[{
   RowBox[{"QAAddReplacementRules", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       SubscriptBox["k_", "0"], "^", "2"}], "\[RuleDelayed]", 
      RowBox[{
       RowBox[{
        SubscriptBox["k", "1"], "^", "2"}], "+", 
       RowBox[{
        SubscriptBox["k", "2"], "^", "2"}], "+", 
       RowBox[{
        SubscriptBox["k", "3"], "^", "2"}]}]}], "}"}], "]"}], ";"}], " ", 
  "*)"}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.7634621407951145`*^9, 3.763462160013775*^9}, {
  3.7634700888529572`*^9, 3.7634701012827053`*^9}, {3.764231964403028*^9, 
  3.764232001487852*^9}, {3.7688842181281347`*^9, 3.768884260465473*^9}, {
  3.7690888712570066`*^9, 3.7690888970709496`*^9}, {3.794226848356415*^9, 
  3.794226852824474*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"QAIntegrate", "[", 
   RowBox[{
    RowBox[{"QAIntegrate", "[", 
     RowBox[{"exp_", ",", "areai_", ",", "ni_Integer", ",", 
      RowBox[{"Sequence", "[", 
       RowBox[{"ivar1_", ",", "ivar2_"}], "]"}]}], "]"}], ",", "areao_", ",", 
    "no_Integer", ",", "ivaro_"}], "]"}], ":=", 
  RowBox[{"QAIntegrate", "[", 
   RowBox[{
    RowBox[{"QAIntegrate", "[", 
     RowBox[{"exp", ",", "areao", ",", "no", ",", "ivaro"}], "]"}], ",", 
    "areai", ",", "ni", ",", "ivar1", ",", "ivar2"}], "]"}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.7633910818789577`*^9, 3.7633911080403156`*^9}, 
   3.763391146228181*^9, {3.7633911988744507`*^9, 3.7633912335208282`*^9}, {
   3.763391273001235*^9, 3.763391389238452*^9}, {3.76339158183239*^9, 
   3.7633917414245663`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{"It", " ", "seems"}], ",", " ", 
    RowBox[{
    "every", " ", "notebook", " ", "must", " ", "use", " ", "its", " ", "own",
      " ", "set", " ", "and", " ", "order", " ", "of", " ", "rules"}]}], 
   "\[IndentingNewLine]", "*)"}], "\n", 
  RowBox[{
   RowBox[{"QAEvaluate2", "[", "ex_", "]"}], ":=", " ", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{
             RowBox[{
              RowBox[{
               RowBox[{
                RowBox[{
                 RowBox[{
                  RowBox[{"ex", "//", 
                   RowBox[{"QAReplace", "[", "\"\<IntegrateJoin\>\"", "]"}]}],
                   "//", 
                  RowBox[{
                  "QAReplaceRepeated", "[", "\"\<CommutatorLocal-Vectors\>\"",
                    "]"}]}], "//", 
                 RowBox[{
                 "QAReplaceRepeated", "[", "\"\<IntegrateSplit\>\"", "]"}]}], 
                "//", 
                RowBox[{
                "QAReplaceRepeated", "[", "\"\<FullDistributionLaw\>\"", 
                 "]"}]}], "//", 
               RowBox[{"QAReplace", "[", "\"\<IntegrateJoin\>\"", "]"}]}], "//", 
              RowBox[{
              "QAReplaceRepeated", "[", "\"\<FullDistributionLaw\>\"", 
               "]"}]}], "//", 
             RowBox[{
              RowBox[{"Factorise", "[", 
               RowBox[{"#", ",", 
                RowBox[{"{", 
                 RowBox[{
                  RowBox[{"Power", "[", 
                   RowBox[{"E", ",", "_"}], "]"}], " ", 
                  SubscriptBox["k", "_"]}], "}"}]}], "]"}], "&"}]}], "//", 
            RowBox[{
            "QAReplace", "[", "\"\<CommutatorLocal-Vectors\>\"", "]"}]}], "//", 
           RowBox[{"QAReplaceRepeated", "[", "\"\<Vectors\>\"", "]"}]}], "//", 
          RowBox[{"QAReplaceRepeated", "[", "\"\<IntegrateSplit\>\"", "]"}]}],
          "//", 
         RowBox[{
         "QAReplaceRepeated", "[", "\"\<FullDistributionLaw\>\"", "]"}]}], "//",
         "QASniff"}], "//", 
       RowBox[{"QAReplace", "[", "\"\<IntegrateJoin\>\"", "]"}]}], "//", 
      RowBox[{"QASimplify", "[", "]"}]}], "//", 
     RowBox[{"QAReplaceRepeated", "[", "\"\<Vectors\>\"", "]"}]}], "//", 
    RowBox[{"QASimplify", "[", "]"}]}]}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.7637035815381756`*^9, 3.7637036018847747`*^9}, {
   3.763704904657044*^9, 3.7637049159857545`*^9}, {3.7637051179647317`*^9, 
   3.7637051478099337`*^9}, 3.7644978929225388`*^9, 3.764498523973798*^9, {
   3.764590507103718*^9, 3.764590509133288*^9}, {3.7647647306959257`*^9, 
   3.764764741831143*^9}, {3.764764775899622*^9, 3.764764856657626*^9}, {
   3.764765167491832*^9, 3.7647651697448053`*^9}, {3.7652989764624968`*^9, 
   3.765298980615385*^9}, 3.765299520942945*^9, {3.768908254005891*^9, 
   3.768908289339499*^9}, 3.76890832500704*^9, {3.769091088445283*^9, 
   3.769091109565817*^9}, 3.7692314885708375`*^9, {3.79481542089876*^9, 
   3.794815489385684*^9}, {3.7948158576602683`*^9, 3.7948158782442107`*^9}, {
   3.794978893331396*^9, 3.7949788960351706`*^9}, 3.7949789279715405`*^9, {
   3.7949908963705144`*^9, 3.7949908985367603`*^9}, {3.7949917135598745`*^9, 
   3.7949917208334293`*^9}, 3.7949917581227503`*^9, {3.7950105142178497`*^9, 
   3.7950105240685196`*^9}, 3.7950107986754613`*^9, 3.7950109406916513`*^9, {
   3.7950111056497245`*^9, 3.795011109926264*^9}, {3.7950608666447506`*^9, 
   3.7950608711237774`*^9}, {3.7950614641474223`*^9, 3.795061480111747*^9}, 
   3.7950621337602262`*^9, {3.7950636677000093`*^9, 3.7950636867630568`*^9}, 
   3.7950650011363583`*^9, {3.795087829666307*^9, 3.795087849365649*^9}, {
   3.7950884666459627`*^9, 3.7950884691432867`*^9}, 3.7950888343625402`*^9, {
   3.79509086573079*^9, 3.795090873330511*^9}, 3.7950909140297394`*^9, {
   3.7950990486376777`*^9, 3.795099060943731*^9}, 3.795099385808359*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Application in QFT - Vectors", "Section",
 CellChangeTimes->{{3.7530956616393967`*^9, 3.753095675373624*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "Define", " ", "the", " ", "exponent", " ", "of", " ", "wave", " ", 
    "function", " ", "solutions"}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"exp", ":=", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      RowBox[{"Sum", "[", 
       RowBox[{
        RowBox[{
         SubscriptBox["k", "i"], " ", 
         SuperscriptBox["x", "i"]}], ",", 
        RowBox[{"{", 
         RowBox[{"i", ",", "0", ",", "3"}], "}"}]}], "]"}]}], "]"}]}], "\n", 
   RowBox[{"(*", " ", 
    RowBox[{"The", " ", "Electromagnetic", " ", "Field", " ", 
     RowBox[{"(", 
      RowBox[{"Vector", " ", "Field"}], ")"}]}], " ", "*)"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    SuperscriptBox["A", "\[Mu]_"], ":=", 
    RowBox[{
     RowBox[{
      SubscriptBox["\[Integral]", "\[Infinity]"], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"1", "/", 
          RowBox[{"Sqrt", "[", 
           RowBox[{"2", " ", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"2", " ", "\[Pi]"}], ")"}], "^", "3"}], " ", 
            SubscriptBox["k", "0"]}], "]"}]}], " ", 
         RowBox[{
          UnderoverscriptBox["\[Sum]", 
           RowBox[{"r", "=", "0"}], "3"], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             SubsuperscriptBox["\[Epsilon]", 
              RowBox[{
               SubscriptBox["k", "1"], ",", 
               SubscriptBox["k", "2"], ",", 
               SubscriptBox["k", "3"]}], 
              RowBox[{"\[Mu]", ",", "r"}]], " ", 
             SubsuperscriptBox[
              OverscriptBox["a", "^"], 
              RowBox[{
               SubscriptBox["k", "1"], ",", 
               SubscriptBox["k", "2"], ",", 
               SubscriptBox["k", "3"]}], "r"]}], ")"}], " ", 
           RowBox[{"Exp", "[", 
            RowBox[{
             RowBox[{"-", "\[ImaginaryI]"}], " ", "exp"}], "]"}]}]}]}], ")"}], 
       RowBox[{
        SuperscriptBox["\[DifferentialD]", "3"], "k"}]}]}], "+", 
     RowBox[{
      SubscriptBox["\[Integral]", "\[Infinity]"], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"1", "/", 
          RowBox[{"Sqrt", "[", 
           RowBox[{"2", " ", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"2", " ", "\[Pi]"}], ")"}], "^", "3"}], " ", 
            SubscriptBox["k", "0"]}], "]"}]}], " ", 
         RowBox[{
          UnderoverscriptBox["\[Sum]", 
           RowBox[{"r", "=", "0"}], "3"], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             SubsuperscriptBox["\[Epsilon]", 
              RowBox[{
               SubscriptBox["k", "1"], ",", 
               SubscriptBox["k", "2"], ",", 
               SubscriptBox["k", "3"]}], 
              RowBox[{"\[Mu]", ",", "r"}]], " ", 
             SubsuperscriptBox[
              OverscriptBox["a", "^"], 
              RowBox[{
               SubscriptBox["k", "1"], ",", 
               SubscriptBox["k", "2"], ",", 
               SubscriptBox["k", "3"]}], 
              RowBox[{"r", ",", "\[Dagger]"}]]}], ")"}], " ", 
           RowBox[{"Exp", "[", 
            RowBox[{"\[ImaginaryI]", " ", "exp"}], "]"}]}]}]}], ")"}], 
       RowBox[{
        SuperscriptBox["\[DifferentialD]", "3"], "k"}]}]}]}]}]}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.746790561237116*^9, 3.746790579017577*^9}, {
   3.746790624828087*^9, 3.7467906832289424`*^9}, {3.7467910806134157`*^9, 
   3.746791097648903*^9}, {3.7467912451504765`*^9, 3.746791359024001*^9}, {
   3.7472853208397675`*^9, 3.747285355043347*^9}, {3.764232474317463*^9, 
   3.7642324930264125`*^9}, 3.7642340791524296`*^9, {3.76440305526606*^9, 
   3.7644030850154767`*^9}, {3.7688842971582737`*^9, 3.768884338894638*^9}, {
   3.7688843836788282`*^9, 3.7688844051463823`*^9}, {3.7688845498942027`*^9, 
   3.7688846516798677`*^9}}],

Cell[CellGroupData[{

Cell["Hamiltonian", "Subsection",
 CellChangeTimes->{{3.753181189848268*^9, 3.753181193541355*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"deriv0", "[", 
   RowBox[{"\[Mu]_", ",", "\[Nu]_"}], "]"}], ":=", 
  RowBox[{
   SubscriptBox["\[GothicCapitalD]", 
    SuperscriptBox["x", "\[Nu]"]], 
   RowBox[{"(", 
    RowBox[{
     SuperscriptBox["A", "\[Mu]"], "//", "Evaluate"}], 
    ")"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"deriv1", "[", 
   RowBox[{"\[Mu]_", ",", "\[Nu]_"}], "]"}], ":=", 
  RowBox[{"Sum", "[", 
   RowBox[{
    RowBox[{
     SuperscriptBox["g", 
      RowBox[{"\[Nu]", ",", "\[Rho]"}]], " ", 
     RowBox[{"deriv0", "[", 
      RowBox[{"\[Mu]", ",", "\[Rho]"}], "]"}]}], ",", 
    RowBox[{"{", 
     RowBox[{"\[Rho]", ",", "0", ",", "3"}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"deriv2", "[", 
   RowBox[{"\[Mu]_", ",", "\[Nu]_"}], "]"}], ":=", 
  RowBox[{"Sum", "[", 
   RowBox[{
    RowBox[{
     SubscriptBox["g", 
      RowBox[{"\[Mu]", ",", "\[Tau]"}]], " ", 
     RowBox[{"deriv1", "[", 
      RowBox[{"\[Tau]", ",", "\[Nu]"}], "]"}]}], ",", 
    RowBox[{"{", 
     RowBox[{"\[Tau]", ",", "0", ",", "3"}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"\[GothicCapitalH]", ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"-", "1"}], "/", "2"}], " ", 
   RowBox[{"Sum", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"deriv2", "[", 
       RowBox[{"\[Mu]", ",", "\[Nu]"}], "]"}], "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"deriv1", "[", 
         RowBox[{"\[Mu]", ",", "\[Nu]"}], "]"}], "/.", 
        RowBox[{"k", "\[Rule]", "k1"}]}], ")"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"\[Mu]", ",", "0", ",", "3"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"\[Nu]", ",", "0", ",", "3"}], "}"}]}], 
    "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{"H", ":=", 
  RowBox[{
   SubscriptBox["\[Integral]", "\[Infinity]"], 
   RowBox[{"\[GothicCapitalH]", 
    RowBox[{
     SuperscriptBox["\[DifferentialD]", "3"], "x"}]}]}]}]}], "Input",
 InitializationCell->True,
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmViYGAQBWIQHfTOk23Bt9eOBtFZYDri1DIFEN2mna8OohtOCxqD
6LQHh0xAtFGFozWIZon+bAeiEyKKXED0I5tKMF1196gniF7kfNYHTEeGBYLo
ikzLEBB97HdMFFh/xsdoEK30/GoSiN6a86kSRNdNW9cEtt8mrxdEX+vbOBtE
ezvZzQXRFtx/V4HolYERq0G0XUHvFhAt4p+xFWzfd9GjINqtNQpMb8h5mlP4
/bXjpL23c0H0q+nZ8q3cbxzX3+9RB9FRfA8MQbTCy7smIDqpd+3ktUD6z8zM
qSCaLUN04xYgHSNrvhlEf3s3+wuINrj94CuIPhDh/h9E8y2+wLQVSAMAGqu8
lw==
  "]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"QAOptions", "[", 
  RowBox[{
   RowBox[{"Verbosity", "\[Rule]", "1"}], ",", 
   RowBox[{"MaxLines", "\[Rule]", "Infinity"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7950133641896353`*^9, 3.795013387221075*^9}, 
   3.7950144986427417`*^9, 3.7950612994895797`*^9, {3.7950616587562065`*^9, 
   3.795061659686753*^9}, {3.795063643311244*^9, 3.795063643474767*^9}, {
   3.7950712012790136`*^9, 3.7950712093654013`*^9}, 3.7950716754280596`*^9, {
   3.795071711720945*^9, 3.7950717224403124`*^9}, 3.795073129780592*^9, 
   3.7950884884416656`*^9, {3.795099089563261*^9, 3.795099089693878*^9}, 
   3.795109811948762*^9, {3.79511396568373*^9, 3.7951139658223195`*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Verbosity", "\[Rule]", "1"}], ",", 
   RowBox[{"MaxLines", "\[Rule]", "\[Infinity]"}], ",", 
   RowBox[{"PrintStack", "\[Rule]", "None"}], ",", 
   RowBox[{"StackPattern", "\[Rule]", "None"}], ",", 
   RowBox[{
   "QuantumAlgebra`Options`Private`LinesAlreadyPrinted", "\[Rule]", "0"}], 
   ",", 
   RowBox[{"DefinitionLevel", "\[Rule]", "3"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7951139667957497`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"hamiltonian", "=", 
  RowBox[{
   RowBox[{"H", "//", "QAEvaluate2"}], "//", "QATiming"}]}]], "Input",
 CellChangeTimes->{{3.7531828032556176`*^9, 3.7531828110986414`*^9}, {
  3.7536121486020994`*^9, 3.7536121637905006`*^9}, {3.764589693203017*^9, 
  3.764589696269812*^9}, {3.764591771005744*^9, 3.7645917750997877`*^9}, {
  3.764764903944191*^9, 3.7647649041376686`*^9}, {3.769091129017764*^9, 
  3.769091130249447*^9}, {3.7949789730576506`*^9, 3.794978973298008*^9}, {
  3.794991034561247*^9, 3.7949910385017138`*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"\<\"Mon 6 Apr 2020 10:12:44\"\>", ",", "174.75`"}], "}"}], ",", 
   RowBox[{
    RowBox[{"2", " ", 
     OverscriptBox["1", "^"], " ", 
     RowBox[{
      SubscriptBox["\[Integral]", "\[Infinity]"], 
      RowBox[{
       SubscriptBox["k", "0"], 
       RowBox[{
        SuperscriptBox["\[DifferentialD]", "3"], "k"}]}]}]}], "-", 
    RowBox[{
     SubscriptBox["\[Integral]", "\[Infinity]"], 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], 
            RowBox[{"0", ",", "\[Dagger]"}]], "\[CenterDot]", 
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], "0"]}], "-", 
          RowBox[{
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], 
            RowBox[{"1", ",", "\[Dagger]"}]], "\[CenterDot]", 
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], "1"]}], "-", 
          RowBox[{
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], 
            RowBox[{"2", ",", "\[Dagger]"}]], "\[CenterDot]", 
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], "2"]}], "-", 
          RowBox[{
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], 
            RowBox[{"3", ",", "\[Dagger]"}]], "\[CenterDot]", 
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], "3"]}]}], ")"}], " ", 
        SubscriptBox["k", "0"]}], ")"}], 
      RowBox[{
       SuperscriptBox["\[DifferentialD]", "3"], "k"}]}]}]}]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.795015122459836*^9, 3.7950175725965776`*^9, 
  3.79506212324234*^9, 3.795062698049274*^9, 3.7950640356209755`*^9, 
  3.7950655280977745`*^9, 3.7950716613257227`*^9, 3.7950719792469125`*^9, 
  3.795073027481552*^9, 3.795074212513315*^9, 3.7950750812805147`*^9, 
  3.795088070141564*^9, 3.795088667911917*^9, 3.795089117260785*^9, 
  3.795091674631892*^9, 3.7950923460419345`*^9, 3.7950992861058383`*^9, 
  3.7950996202921143`*^9, 3.7951083896310883`*^9, 3.795110400790847*^9, 
  3.7951149785720387`*^9, 3.79514832767352*^9, 3.7951497525282006`*^9},
 Background->RGBColor[1, 1, 0.85]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Elaboration of Simplification Chain", "Subsection",
 CellChangeTimes->{{3.7647358980787015`*^9, 3.7647359133707933`*^9}}],

Cell[CellGroupData[{

Cell["Initial Calculation", "Subsubsection",
 CellChangeTimes->{{3.7652976846027355`*^9, 3.7652976967223144`*^9}}],

Cell[BoxData[
 RowBox[{"h1", "=", 
  RowBox[{"H", "//", "Timing"}]}]], "Input",
 CellChangeTimes->{
  3.7647359335787315`*^9, {3.764736007328117*^9, 3.7647360087423334`*^9}, {
   3.764737360322072*^9, 3.764737364390156*^9}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Join", "Subsubsection",
 CellChangeTimes->{{3.765297729704084*^9, 3.765297731650878*^9}, {
  3.794991153993739*^9, 3.794991158468776*^9}}],

Cell[BoxData[
 RowBox[{"h2", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"h1", "[", 
     RowBox[{"[", "2", "]"}], "]"}], "//", 
    RowBox[{"QAReplace", "[", "\"\<IntegrateJoin\>\"", "]"}]}], "//", 
   "Timing"}]}]], "Input",
 CellChangeTimes->{{3.7647360758782096`*^9, 3.7647360805446987`*^9}, {
   3.764737391432677*^9, 3.764737401466857*^9}, 3.7650398401209702`*^9, 
   3.7650987128425226`*^9, 3.7946718942521906`*^9, {3.794755078794705*^9, 
   3.7947550813628078`*^9}, 3.794755476158491*^9, {3.794764132394354*^9, 
   3.7947641464597564`*^9}, {3.7947645464966497`*^9, 3.794764560641837*^9}, {
   3.794764598159548*^9, 3.794764668517471*^9}, 3.794764704141245*^9, {
   3.7947656248479223`*^9, 3.7947656344113626`*^9}, 3.7949789010737*^9}],

Cell["\<\
Goal: Join integrals
Result: The integration order is exchanged\
\>", "Text",
 CellChangeTimes->{{3.7647368027337227`*^9, 3.7647368890658064`*^9}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Apply Commutators", "Subsubsection",
 CellChangeTimes->{{3.7652978117321944`*^9, 3.765297820617427*^9}, {
   3.7942312575232244`*^9, 3.7942312818023825`*^9}, 3.794237739391265*^9, {
   3.7949911649235225`*^9, 3.7949911703849235`*^9}}],

Cell[BoxData[
 RowBox[{"h3", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"h2", "[", 
     RowBox[{"[", "2", "]"}], "]"}], "//", 
    RowBox[{"QAReplace", "[", "\"\<CommutatorLocal-Vectors\>\"", "]"}]}], "//",
    "Timing"}]}]], "Input",
 CellChangeTimes->{{3.7647361735290737`*^9, 3.764736201021571*^9}, {
   3.7647374169932966`*^9, 3.764737422609273*^9}, 3.765098763074263*^9, 
   3.7689063486574993`*^9, 3.7692529300477843`*^9, {3.794231053177472*^9, 
   3.794231080513405*^9}, 3.7942314720568266`*^9, {3.7944583043637085`*^9, 
   3.794458306026265*^9}, 3.7946646533511877`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Split, DistributionLaw", "Subsubsection",
 CellChangeTimes->{{3.765297905039586*^9, 3.7652979070651703`*^9}, {
  3.7942378180954523`*^9, 3.7942378199115973`*^9}, {3.7949911779207783`*^9, 
  3.7949911963448915`*^9}}],

Cell[BoxData[
 RowBox[{"h5", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"h3", "[", 
     RowBox[{"[", "2", "]"}], "]"}], "//", 
    RowBox[{"QAReplaceRepeated", "[", "\"\<IntegrateSplit\>\"", "]"}]}], "//",
    "Timing"}]}]], "Input",
 CellChangeTimes->{{3.7651074494007945`*^9, 3.765107476193163*^9}, 
   3.7652978887641263`*^9, {3.769252988364731*^9, 3.7692530090693398`*^9}, {
   3.7692627018137026`*^9, 3.7692627288403983`*^9}, {3.7692632483056307`*^9, 
   3.769263250295337*^9}, {3.794237948634473*^9, 3.7942379855812163`*^9}, 
   3.7942847356634827`*^9, {3.794632765438269*^9, 3.7946327701077766`*^9}, {
   3.7946508054633293`*^9, 3.794650835703492*^9}, 3.794672127288561*^9, {
   3.7946723957958126`*^9, 3.7946723977039833`*^9}, 3.794815661273202*^9, 
   3.7949789292252154`*^9}],

Cell[BoxData[
 RowBox[{"h6", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"h5", "[", 
     RowBox[{"[", "2", "]"}], "]"}], "//", 
    RowBox[{"QAReplaceRepeated", "[", "\"\<FullDistributionLaw\>\"", "]"}]}], 
   "//", "Timing"}]}]], "Input",
 CellChangeTimes->{{3.7946508440282393`*^9, 3.7946508529982615`*^9}, {
  3.7947241747904015`*^9, 3.7947241929718375`*^9}}],

Cell["\<\
Volume integrals are resolved, double integrals are resolved.\
\>", "Text",
 CellChangeTimes->{{3.7652979995537863`*^9, 3.7652980304780235`*^9}}],

Cell["\<\
Goal: Simplify single integrals
Result: Application of distribution law lets extract constant factors and \
permits further simplification like resolution of volume integrals and double \
integrals. Only one not nested integral from 3 nested ones remains.\
\>", "Text",
 CellChangeTimes->{{3.764737060760664*^9, 3.764737168584217*^9}, {
   3.764737281179407*^9, 3.764737316970703*^9}, {3.764737523979766*^9, 
   3.764737526868039*^9}, {3.7647375917604475`*^9, 3.7647376371530504`*^9}, 
   3.7654294175429697`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Join, DistributionLaw", "Subsubsection",
 CellChangeTimes->{{3.765298552242305*^9, 3.765298571746129*^9}, {
  3.794238740105754*^9, 3.7942387430578637`*^9}, {3.794238825220231*^9, 
  3.7942388288026857`*^9}, {3.7949912043365035`*^9, 3.7949912202968397`*^9}}],

Cell[BoxData[
 RowBox[{"h7", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"h6", "[", 
     RowBox[{"[", "2", "]"}], "]"}], "//", 
    RowBox[{"QAReplace", "[", "\"\<IntegrateJoin\>\"", "]"}]}], "//", 
   "Timing"}]}]], "Input",
 CellChangeTimes->{{3.764736671142496*^9, 3.764736675508848*^9}, {
   3.7647374639189315`*^9, 3.764737469771239*^9}, 3.765082561002779*^9, {
   3.765095165797535*^9, 3.765095167790205*^9}, {3.7650989471528187`*^9, 
   3.765098978578753*^9}, {3.7652767267507124`*^9, 3.7652767333011904`*^9}, {
   3.7652981375665503`*^9, 3.7652981426130495`*^9}, {3.768907053270625*^9, 
   3.7689070930780907`*^9}, {3.769263629838964*^9, 3.769263639440281*^9}, {
   3.7692637918236156`*^9, 3.76926380165631*^9}, 3.769264306944508*^9, 
   3.7693338620588984`*^9, 3.794210121973859*^9, 3.7942388113493094`*^9, {
   3.794284777591403*^9, 3.794284805798479*^9}, 3.7942866850423355`*^9, {
   3.7942869711126337`*^9, 3.7942869754689827`*^9}, {3.7945651703142495`*^9, 
   3.7945651723089237`*^9}, 3.7946362666513667`*^9, {3.7946509258535547`*^9, 
   3.794650935188602*^9}, 3.794672175881668*^9, {3.7946732824348173`*^9, 
   3.7946732845890923`*^9}, 3.7947182714931355`*^9, {3.7947183790711465`*^9, 
   3.7947183810638175`*^9}, 3.7947240233836403`*^9, {3.794736667345191*^9, 
   3.794736669313893*^9}, 3.794753957075539*^9, {3.7947559950913134`*^9, 
   3.7947560006394453`*^9}, 3.7947659524454365`*^9, 3.7949789037286024`*^9}],

Cell[BoxData[
 RowBox[{"h8", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"h7", "[", 
     RowBox[{"[", "2", "]"}], "]"}], "//", 
    RowBox[{"QAReplaceRepeated", "[", "\"\<FullDistributionLaw\>\"", "]"}]}], 
   "//", "Timing"}]}]], "Input",
 CellChangeTimes->{{3.7946508440282393`*^9, 3.7946508529982615`*^9}, 
   3.79465091089554*^9, {3.7946509519258595`*^9, 3.794650963309429*^9}}],

Cell["\<\
Goal: Apply contraction
Result: Outstanding simplifications\
\>", "Text",
 CellChangeTimes->{{3.7647379264876556`*^9, 3.7647380497628403`*^9}, {
  3.794239293686801*^9, 3.794239297636243*^9}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Simplification", "Subsubsection",
 CellChangeTimes->{{3.765298712780841*^9, 3.765298717652809*^9}, {
   3.7652988497055545`*^9, 3.7652988522697363`*^9}, 3.794239592052211*^9}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"h9", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"h8", "[", 
         RowBox[{"[", "2", "]"}], "]"}], "//", 
        RowBox[{"QAReplace", "[", "\"\<Vectors\>\"", "]"}]}], "//", 
       RowBox[{"QASimplify", "[", "]"}]}], "//", 
      RowBox[{"QAReplace", "[", "\"\<Vectors\>\"", "]"}]}], "//", 
     RowBox[{"QASimplify", "[", "]"}]}], "//", "Timing"}]}], " ", 
  "\[IndentingNewLine]"}]], "Input",
 CellChangeTimes->{{3.7647381991506815`*^9, 3.764738253730645*^9}, {
   3.7647387640014753`*^9, 3.764738765896406*^9}, 3.764763724524659*^9, 
   3.7647639171623273`*^9, {3.764764046237031*^9, 3.7647640645460525`*^9}, {
   3.764764395196371*^9, 3.7647644001222*^9}, {3.7649404906048965`*^9, 
   3.764940492538717*^9}, {3.7649986582734137`*^9, 3.7649986710861764`*^9}, 
   3.765000087429845*^9, {3.765000289409845*^9, 3.765000298759842*^9}, {
   3.7650196851329765`*^9, 3.765019694748255*^9}, 3.765019865254609*^9, 
   3.7650203603146987`*^9, 3.7650318987569637`*^9, {3.7652769360124807`*^9, 
   3.7652769361281347`*^9}, {3.765276971057694*^9, 3.765276973599891*^9}, 
   3.7652773580808516`*^9, {3.76529861080165*^9, 3.765298618598792*^9}, 
   3.7689071898851066`*^9, 3.7692534819215107`*^9, 3.7692542648533134`*^9, 
   3.769257339671487*^9, {3.769257417714718*^9, 3.76925743789872*^9}, 
   3.769258089592368*^9, {3.7693339146958985`*^9, 3.769333914822853*^9}, 
   3.769334066895273*^9, 3.769335943658669*^9, {3.7942304044287043`*^9, 
   3.7942304109552727`*^9}, {3.794239555438574*^9, 3.79423956957031*^9}, {
   3.7942440781003137`*^9, 3.794244084645788*^9}, {3.794246532861232*^9, 
   3.79424653378975*^9}, {3.7942871356009436`*^9, 3.7942871387634697`*^9}, {
   3.794287975368723*^9, 3.7942879958843937`*^9}, {3.7946509878688984`*^9, 
   3.7946509909486656`*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"90.703125`", ",", 
   RowBox[{
    RowBox[{"2", " ", 
     OverscriptBox["1", "^"], " ", 
     RowBox[{
      SubscriptBox["\[Integral]", "\[Infinity]"], 
      RowBox[{
       SubscriptBox["k", "0"], 
       RowBox[{
        SuperscriptBox["\[DifferentialD]", "3"], "k"}]}]}]}], "-", 
    RowBox[{
     SubscriptBox["\[Integral]", "\[Infinity]"], 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], 
            RowBox[{"0", ",", "\[Dagger]"}]], "\[CenterDot]", 
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], "0"]}], "-", 
          RowBox[{
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], 
            RowBox[{"1", ",", "\[Dagger]"}]], "\[CenterDot]", 
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], "1"]}], "-", 
          RowBox[{
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], 
            RowBox[{"2", ",", "\[Dagger]"}]], "\[CenterDot]", 
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], "2"]}], "-", 
          RowBox[{
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], 
            RowBox[{"3", ",", "\[Dagger]"}]], "\[CenterDot]", 
           SubsuperscriptBox[
            OverscriptBox["a", "^"], 
            RowBox[{
             SubscriptBox["k", "1"], ",", 
             SubscriptBox["k", "2"], ",", 
             SubscriptBox["k", "3"]}], "3"]}]}], ")"}], " ", 
        SubscriptBox["k", "0"]}], ")"}], 
      RowBox[{
       SuperscriptBox["\[DifferentialD]", "3"], "k"}]}]}]}]}], 
  "}"}]], "Output",
 CellChangeTimes->CompressedData["
1:eJwVxU0og3EcAOAptTX7CFutHNaWGiNkZMjHm2FDatSUZZhJQ20OStqMlZLV
Xopy8JXVHJCNhVDU5mBiN5GVltD4JyFfOfj9Dk+PxGhp6k5gMBgK8Pt6qljj
EMp7ljaAB7RF9C5saBlaxTNjDWd4zuD6Bz4j+fJzuYRiiwQPuNHjucmAo3OK
DDlstkdqcK7LdIcHt1O/8SQ/zcuCyS2XLoBZI1YP/rcxFsMTSSh9Cy7lVxTj
6vnoxSGs2yw8XhQQqjFilS/B6j3Hda2QUKbz+2I13CruE2vg3oOFAH6p1YVx
o2k6tQ6eeMxTWmDl5o/UCpd7S0IxFaHCplxDHGbkW21PsMw8RT/D+/qeFbxZ
4ZO8wh33rso3OCQrm3qH7W7bMs55iauY1YTaY6ud+CTfuY2PvmlIMjwc+EzW
wyqZI9gGC331cTfcJeusouHZK17/DswcGGeF4BShqP0EPpJmDzJrCPUPyWq9
qA==
  "]]
}, Open  ]],

Cell["\<\
Goal: Apply substitution to eliminate polarization vectors
Result: Done.\
\>", "Text",
 CellChangeTimes->{{3.7647390539951086`*^9, 3.7647391239669037`*^9}, {
  3.7647643604557266`*^9, 3.7647643632542615`*^9}}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["3-Momentum", "Subsection",
 CellChangeTimes->{{3.764136688256733*^9, 3.764136691932936*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[GothicP]", "[", "i_", "]"}], ":=", 
  RowBox[{"-", 
   RowBox[{"Sum", "[", 
    RowBox[{
     RowBox[{
      SubscriptBox["g", 
       RowBox[{"\[Mu]", ",", "\[Nu]"}]], " ", 
      RowBox[{
       RowBox[{"deriv0", "[", 
        RowBox[{"\[Mu]", ",", "0"}], "]"}], "\[CenterDot]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"deriv0", "[", 
          RowBox[{"\[Nu]", ",", "i"}], "]"}], "/.", 
         RowBox[{"k", "\[Rule]", "k1"}]}], ")"}]}]}], ",", 
     RowBox[{"{", 
      RowBox[{"\[Mu]", ",", "0", ",", "3"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"\[Nu]", ",", "0", ",", "3"}], "}"}]}], 
    "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"p", "[", "i_", "]"}], ":=", 
  RowBox[{
   SubscriptBox["\[Integral]", "\[Infinity]"], 
   RowBox[{
    RowBox[{"\[GothicP]", "[", "i", "]"}], 
    RowBox[{
     SuperscriptBox["\[DifferentialD]", "3"], "x"}]}]}]}]}], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.7641367019750657`*^9, 3.764136837580315*^9}, {
  3.7641368828461857`*^9, 3.7641369261138277`*^9}, {3.7641369637870483`*^9, 
  3.7641370048731747`*^9}, {3.7642163294658937`*^9, 3.764216359648203*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"momentum", "=", 
  RowBox[{
   RowBox[{"Map", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"QAEvaluate2", "[", 
       RowBox[{"p", "[", "#1", "]"}], "]"}], "&"}], ",", 
     RowBox[{"Range", "[", "3", "]"}]}], "]"}], "//", "QATiming"}]}]], "Input",
 CellChangeTimes->{{3.764216524336418*^9, 3.764216555885026*^9}, {
   3.7642168754860587`*^9, 3.7642168906438785`*^9}, {3.7642169465114636`*^9, 
   3.764216960933875*^9}, 3.7644035404250803`*^9, {3.764497916721094*^9, 
   3.764497917309475*^9}, 3.7645089264163823`*^9, 3.764845335944611*^9, 
   3.7648457560678873`*^9, {3.7949794982703686`*^9, 3.794979498627383*^9}, {
   3.7949913129916515`*^9, 3.7949913150252137`*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"\<\"Mon 6 Apr 2020 10:16:33\"\>", ",", "64.3125`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      SubscriptBox["\[Integral]", "\[Infinity]"], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"(", 
             RowBox[{
              SubsuperscriptBox[
               OverscriptBox["a", "^"], 
               RowBox[{
                SubscriptBox["k", "1"], ",", 
                SubscriptBox["k", "2"], ",", 
                SubscriptBox["k", "3"]}], 
               RowBox[{"0", ",", "\[Dagger]"}]], "\[CenterDot]", 
              SubsuperscriptBox[
               OverscriptBox["a", "^"], 
               RowBox[{
                SubscriptBox["k", "1"], ",", 
                SubscriptBox["k", "2"], ",", 
                SubscriptBox["k", "3"]}], "0"]}], ")"}]}], "+", 
           RowBox[{
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], 
             RowBox[{"1", ",", "\[Dagger]"}]], "\[CenterDot]", 
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], "1"]}], "+", 
           RowBox[{
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], 
             RowBox[{"2", ",", "\[Dagger]"}]], "\[CenterDot]", 
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], "2"]}], "+", 
           RowBox[{
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], 
             RowBox[{"3", ",", "\[Dagger]"}]], "\[CenterDot]", 
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], "3"]}]}], ")"}], " ", 
         SubscriptBox["k", "1"]}], ")"}], 
       RowBox[{
        SuperscriptBox["\[DifferentialD]", "3"], "k"}]}]}], ",", 
     RowBox[{
      SubscriptBox["\[Integral]", "\[Infinity]"], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"(", 
             RowBox[{
              SubsuperscriptBox[
               OverscriptBox["a", "^"], 
               RowBox[{
                SubscriptBox["k", "1"], ",", 
                SubscriptBox["k", "2"], ",", 
                SubscriptBox["k", "3"]}], 
               RowBox[{"0", ",", "\[Dagger]"}]], "\[CenterDot]", 
              SubsuperscriptBox[
               OverscriptBox["a", "^"], 
               RowBox[{
                SubscriptBox["k", "1"], ",", 
                SubscriptBox["k", "2"], ",", 
                SubscriptBox["k", "3"]}], "0"]}], ")"}]}], "+", 
           RowBox[{
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], 
             RowBox[{"1", ",", "\[Dagger]"}]], "\[CenterDot]", 
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], "1"]}], "+", 
           RowBox[{
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], 
             RowBox[{"2", ",", "\[Dagger]"}]], "\[CenterDot]", 
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], "2"]}], "+", 
           RowBox[{
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], 
             RowBox[{"3", ",", "\[Dagger]"}]], "\[CenterDot]", 
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], "3"]}]}], ")"}], " ", 
         SubscriptBox["k", "2"]}], ")"}], 
       RowBox[{
        SuperscriptBox["\[DifferentialD]", "3"], "k"}]}]}], ",", 
     RowBox[{
      SubscriptBox["\[Integral]", "\[Infinity]"], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"(", 
             RowBox[{
              SubsuperscriptBox[
               OverscriptBox["a", "^"], 
               RowBox[{
                SubscriptBox["k", "1"], ",", 
                SubscriptBox["k", "2"], ",", 
                SubscriptBox["k", "3"]}], 
               RowBox[{"0", ",", "\[Dagger]"}]], "\[CenterDot]", 
              SubsuperscriptBox[
               OverscriptBox["a", "^"], 
               RowBox[{
                SubscriptBox["k", "1"], ",", 
                SubscriptBox["k", "2"], ",", 
                SubscriptBox["k", "3"]}], "0"]}], ")"}]}], "+", 
           RowBox[{
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], 
             RowBox[{"1", ",", "\[Dagger]"}]], "\[CenterDot]", 
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], "1"]}], "+", 
           RowBox[{
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], 
             RowBox[{"2", ",", "\[Dagger]"}]], "\[CenterDot]", 
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], "2"]}], "+", 
           RowBox[{
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], 
             RowBox[{"3", ",", "\[Dagger]"}]], "\[CenterDot]", 
            SubsuperscriptBox[
             OverscriptBox["a", "^"], 
             RowBox[{
              SubscriptBox["k", "1"], ",", 
              SubscriptBox["k", "2"], ",", 
              SubscriptBox["k", "3"]}], "3"]}]}], ")"}], " ", 
         SubscriptBox["k", "3"]}], ")"}], 
       RowBox[{
        SuperscriptBox["\[DifferentialD]", "3"], "k"}]}]}]}], "}"}]}], 
  "}"}]], "Output",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQBuJjz7fHruF+4zjjYEYSmF4T0gmiO2fGfwLRJyIsVrPz
vHFsvut3E0S7T3TR4APSs9N60kG0neWrOwJAetrMjWy6QHq1smGZKZBuCU2+
A6IDTi4QNQfSfhP2s1gA6VvVoczWQPqKHkstiC7Im2VhC6TfbVHgsAPS7JP6
9ZKAdBjXpr4dQHrBEs2c+0BaS15jPYh+2qcXspf3jePmDfcjwvjeOGqUhvTO
F3njeCPgStFeEG3dufs7kLaZVmD+A0jH+DfFgehPAv7ywqJvHI87h2zgdHvj
yH/CNCESSG/PklaNB9J7I8vTQbQrH0dPBpAO2eO9AkTf5HmxC0S3Ld7xDkxz
Nv5sANJfI3kCGoH0k7KKNBDds2gG/xQgnX6Qo3UNkAYAgNmc6w==
  "],
 Background->RGBColor[1, 1, 0.85]]
}, Open  ]],

Cell["\<\
The 3-Momentum is different from the Hamiltonian in that it requires \
additional simplification. This is due to the fact that there are some \
additional terms with anti-symmetric integrands. These terms disappear after \
simplification because their integral results in null.\
\>", "Text",
 CellChangeTimes->{{3.7648460487195597`*^9, 3.7648461571225686`*^9}, {
   3.764846194794821*^9, 3.7648462703227754`*^9}, {3.7654292895237713`*^9, 
   3.7654292913559003`*^9}, 3.7654294329090695`*^9}]
}, Open  ]],

Cell["Tests", "Subsection",
 CellChangeTimes->{{3.7643396380727463`*^9, 3.7643396388367414`*^9}}]
}, Open  ]]
}, Open  ]]
},
WindowSize->{838, 879},
WindowMargins->{{418, Automatic}, {Automatic, 96}},
PrintingPageRange->{Automatic, Automatic},
InputAliases->{"com" -> SubscriptBox[
    RowBox[{"[", "\[SelectionPlaceholder]", "]"}], "-"], "op" -> 
  OverscriptBox["\[SelectionPlaceholder]", "^"], "hop" -> SuperscriptBox[
    OverscriptBox["\[SelectionPlaceholder]", "^"], "\[Dagger]"], "opd" -> 
  SubscriptBox[
    OverscriptBox["\[SelectionPlaceholder]", "^"], "\[SelectionPlaceholder]"],
   "opu" -> SuperscriptBox[
    OverscriptBox["\[SelectionPlaceholder]", "^"], "\[SelectionPlaceholder]"],
   "opdu" -> SubsuperscriptBox[
    OverscriptBox["\[SelectionPlaceholder]", "^"], "\[SelectionPlaceholder]", 
    "\[SelectionPlaceholder]"], "hopd" -> SubsuperscriptBox[
    OverscriptBox["\[SelectionPlaceholder]", "^"], "\[SelectionPlaceholder]", 
    "\[Dagger]"], "hopu" -> SuperscriptBox[
    OverscriptBox["\[SelectionPlaceholder]", "^"], 
    RowBox[{"\[SelectionPlaceholder]", ",", "\[Dagger]"}]], "hopdu" -> 
  SubsuperscriptBox[
    OverscriptBox["\[SelectionPlaceholder]", "^"], "\[SelectionPlaceholder]", 
    RowBox[{"\[SelectionPlaceholder]", ",", "\[Dagger]"}]], "ket" -> TagBox[
    RowBox[{
      AdjustmentBox[
      "\[VerticalSeparator]", BoxMargins -> {{-0.4, 0}, {0, 0}}], 
      AdjustmentBox[
       TagBox[
       "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`KetArgs, 
        BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
      "\[RightAngleBracket]"}], QuantumAlgebra`Ket, BaseStyle -> 
    "KetWrapper"], "bra" -> TagBox[
    RowBox[{"\[LeftAngleBracket]", 
      AdjustmentBox[
       TagBox[
       "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
        BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
      AdjustmentBox[
      "\[VerticalSeparator]", BoxMargins -> {{0, -0.2}, {0, 0}}]}], 
    QuantumAlgebra`Bra, BaseStyle -> "BraWrapper"], "braket" -> TagBox[
    RowBox[{"\[LeftAngleBracket]", 
      AdjustmentBox[
       RowBox[{
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraKetArgs,
           BaseStyle -> "BraKetArg"], "\[VerticalSeparator]", 
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraKetArgs,
           BaseStyle -> "BraKetArg"]}], BoxBaselineShift -> 0], 
      "\[RightAngleBracket]"}], QuantumAlgebra`BraKet, BaseStyle -> 
    "BraKetWrapper"], "ketd" -> TagBox[
    SubscriptBox[
     RowBox[{
       AdjustmentBox[
       "\[VerticalSeparator]", BoxMargins -> {{-0.4, 0}, {0, 0}}], 
       AdjustmentBox[
        TagBox[
        "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`KetArgs, 
         BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
       "\[RightAngleBracket]"}], 
     TagBox[
      AdjustmentBox[
       TagBox[
       "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`KetArgs, 
        BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
      QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]], 
    QuantumAlgebra`Ket, BaseStyle -> "KetWrapper"], "ketu" -> TagBox[
    SuperscriptBox[
     RowBox[{
       AdjustmentBox[
       "\[VerticalSeparator]", BoxMargins -> {{-0.4, 0}, {0, 0}}], 
       AdjustmentBox[
        TagBox[
        "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`KetArgs, 
         BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
       "\[RightAngleBracket]"}], 
     TagBox[
      AdjustmentBox[
       TagBox[
       "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`KetArgs, 
        BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
      QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]], 
    QuantumAlgebra`Ket, BaseStyle -> "KetWrapper"], "ketdu" -> TagBox[
    SubsuperscriptBox[
     RowBox[{
       AdjustmentBox[
       "\[VerticalSeparator]", BoxMargins -> {{-0.4, 0}, {0, 0}}], 
       AdjustmentBox[
        TagBox[
        "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`KetArgs, 
         BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
       "\[RightAngleBracket]"}], 
     TagBox[
      AdjustmentBox[
       TagBox[
       "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`KetArgs, 
        BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
      QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"], 
     TagBox[
      AdjustmentBox[
       TagBox[
       "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`KetArgs, 
        BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
      QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]], 
    QuantumAlgebra`Ket, BaseStyle -> "KetWrapper"], "dbra" -> TagBox[
    RowBox[{
      SubscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      RowBox[{"\[LeftAngleBracket]", 
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        AdjustmentBox[
        "\[VerticalSeparator]", BoxMargins -> {{0, -0.2}, {0, 0}}]}]}], 
    QuantumAlgebra`Bra, BaseStyle -> "BraWrapper"], "ubra" -> TagBox[
    RowBox[{
      SuperscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      RowBox[{"\[LeftAngleBracket]", 
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        AdjustmentBox[
        "\[VerticalSeparator]", BoxMargins -> {{0, -0.2}, {0, 0}}]}]}], 
    QuantumAlgebra`Bra, BaseStyle -> "BraWrapper"], "dubra" -> TagBox[
    RowBox[{
      SubsuperscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      RowBox[{"\[LeftAngleBracket]", 
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        AdjustmentBox[
        "\[VerticalSeparator]", BoxMargins -> {{0, -0.2}, {0, 0}}]}]}], 
    QuantumAlgebra`Bra, BaseStyle -> "BraWrapper"], "braketu" -> TagBox[
    SuperscriptBox[
     RowBox[{"\[LeftAngleBracket]", 
       AdjustmentBox[
        RowBox[{
          TagBox[
          "\[SelectionPlaceholder]", 
           QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
           "BraKetArg"], "\[VerticalSeparator]", 
          TagBox[
          "\[SelectionPlaceholder]", 
           QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
           "BraKetArg"]}], BoxBaselineShift -> 0], "\[RightAngleBracket]"}], 
     TagBox[
      AdjustmentBox[
       TagBox[
       "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`KetArgs, 
        BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
      QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]], 
    QuantumAlgebra`BraKet, BaseStyle -> "BraKetWrapper"], "braketdu" -> TagBox[
    SubsuperscriptBox[
     RowBox[{"\[LeftAngleBracket]", 
       AdjustmentBox[
        RowBox[{
          TagBox[
          "\[SelectionPlaceholder]", 
           QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
           "BraKetArg"], "\[VerticalSeparator]", 
          TagBox[
          "\[SelectionPlaceholder]", 
           QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
           "BraKetArg"]}], BoxBaselineShift -> 0], "\[RightAngleBracket]"}], 
     TagBox[
      AdjustmentBox[
       TagBox[
       "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`KetArgs, 
        BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
      QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"], 
     TagBox[
      AdjustmentBox[
       TagBox[
       "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`KetArgs, 
        BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
      QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]], 
    QuantumAlgebra`BraKet, BaseStyle -> "BraKetWrapper"], "dbraket" -> TagBox[
    RowBox[{
      SubscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      RowBox[{"\[LeftAngleBracket]", 
        AdjustmentBox[
         RowBox[{
           TagBox[
           "\[SelectionPlaceholder]", 
            QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
            "BraKetArg"], "\[VerticalSeparator]", 
           TagBox[
           "\[SelectionPlaceholder]", 
            QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
            "BraKetArg"]}], BoxBaselineShift -> 0], 
        "\[RightAngleBracket]"}]}], QuantumAlgebra`BraKet, BaseStyle -> 
    "BraKetWrapper"], "ubraket" -> TagBox[
    RowBox[{
      SuperscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      RowBox[{"\[LeftAngleBracket]", 
        AdjustmentBox[
         RowBox[{
           TagBox[
           "\[SelectionPlaceholder]", 
            QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
            "BraKetArg"], "\[VerticalSeparator]", 
           TagBox[
           "\[SelectionPlaceholder]", 
            QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
            "BraKetArg"]}], BoxBaselineShift -> 0], 
        "\[RightAngleBracket]"}]}], QuantumAlgebra`BraKet, BaseStyle -> 
    "BraKetWrapper"], "dubraket" -> TagBox[
    RowBox[{
      SubsuperscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      RowBox[{"\[LeftAngleBracket]", 
        AdjustmentBox[
         RowBox[{
           TagBox[
           "\[SelectionPlaceholder]", 
            QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
            "BraKetArg"], "\[VerticalSeparator]", 
           TagBox[
           "\[SelectionPlaceholder]", 
            QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
            "BraKetArg"]}], BoxBaselineShift -> 0], 
        "\[RightAngleBracket]"}]}], QuantumAlgebra`BraKet, BaseStyle -> 
    "BraKetWrapper"], "dbraketd" -> TagBox[
    RowBox[{
      SubscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      SubscriptBox[
       RowBox[{"\[LeftAngleBracket]", 
         AdjustmentBox[
          RowBox[{
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"], "\[VerticalSeparator]", 
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"]}], BoxBaselineShift -> 0], "\[RightAngleBracket]"}], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]]}], 
    QuantumAlgebra`BraKet, BaseStyle -> "BraKetWrapper"], "dbraketu" -> TagBox[
    RowBox[{
      SubscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      SuperscriptBox[
       RowBox[{"\[LeftAngleBracket]", 
         AdjustmentBox[
          RowBox[{
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"], "\[VerticalSeparator]", 
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"]}], BoxBaselineShift -> 0], "\[RightAngleBracket]"}], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]]}], 
    QuantumAlgebra`BraKet, BaseStyle -> "BraKetWrapper"], "dbraketdu" -> 
  TagBox[
    RowBox[{
      SubscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      SubsuperscriptBox[
       RowBox[{"\[LeftAngleBracket]", 
         AdjustmentBox[
          RowBox[{
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"], "\[VerticalSeparator]", 
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"]}], BoxBaselineShift -> 0], "\[RightAngleBracket]"}], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]]}], 
    QuantumAlgebra`BraKet, BaseStyle -> "BraKetWrapper"], "ubraketd" -> TagBox[
    RowBox[{
      SuperscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      SubscriptBox[
       RowBox[{"\[LeftAngleBracket]", 
         AdjustmentBox[
          RowBox[{
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"], "\[VerticalSeparator]", 
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"]}], BoxBaselineShift -> 0], "\[RightAngleBracket]"}], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]]}], 
    QuantumAlgebra`BraKet, BaseStyle -> "BraKetWrapper"], "ubraketu" -> TagBox[
    RowBox[{
      SuperscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      SuperscriptBox[
       RowBox[{"\[LeftAngleBracket]", 
         AdjustmentBox[
          RowBox[{
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"], "\[VerticalSeparator]", 
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"]}], BoxBaselineShift -> 0], "\[RightAngleBracket]"}], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]]}], 
    QuantumAlgebra`BraKet, BaseStyle -> "BraKetWrapper"], "ubraketdu" -> 
  TagBox[
    RowBox[{
      SuperscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      SubsuperscriptBox[
       RowBox[{"\[LeftAngleBracket]", 
         AdjustmentBox[
          RowBox[{
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"], "\[VerticalSeparator]", 
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"]}], BoxBaselineShift -> 0], "\[RightAngleBracket]"}], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]]}], 
    QuantumAlgebra`BraKet, BaseStyle -> "BraKetWrapper"], "dubraketd" -> 
  TagBox[
    RowBox[{
      SubsuperscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      SubscriptBox[
       RowBox[{"\[LeftAngleBracket]", 
         AdjustmentBox[
          RowBox[{
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"], "\[VerticalSeparator]", 
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"]}], BoxBaselineShift -> 0], "\[RightAngleBracket]"}], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]]}], 
    QuantumAlgebra`BraKet, BaseStyle -> "BraKetWrapper"], "dubraketu" -> 
  TagBox[
    RowBox[{
      SubsuperscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      SuperscriptBox[
       RowBox[{"\[LeftAngleBracket]", 
         AdjustmentBox[
          RowBox[{
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"], "\[VerticalSeparator]", 
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"]}], BoxBaselineShift -> 0], "\[RightAngleBracket]"}], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]]}], 
    QuantumAlgebra`BraKet, BaseStyle -> "BraKetWrapper"], "dubraketdu" -> 
  TagBox[
    RowBox[{
      SubsuperscriptBox["\[InvisiblePrefixScriptBase]", 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "LIndex"]], 
      SubsuperscriptBox[
       RowBox[{"\[LeftAngleBracket]", 
         AdjustmentBox[
          RowBox[{
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"], "\[VerticalSeparator]", 
            TagBox[
            "\[SelectionPlaceholder]", 
             QuantumAlgebra`QuantumNotation`BraKetArgs, BaseStyle -> 
             "BraKetArg"]}], BoxBaselineShift -> 0], "\[RightAngleBracket]"}], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"], 
       TagBox[
        AdjustmentBox[
         TagBox[
         "\[SelectionPlaceholder]", QuantumAlgebra`QuantumNotation`BraArgs, 
          BaseStyle -> "BraKetArg"], BoxBaselineShift -> 0], 
        QuantumAlgebra`QuantumNotation`UDScript, BaseStyle -> "RIndex"]]}], 
    QuantumAlgebra`BraKet, BaseStyle -> "BraKetWrapper"], "sd" -> 
  SubscriptBox["\[Placeholder]", "\[Placeholder]"], "su" -> 
  SuperscriptBox["\[Placeholder]", "\[Placeholder]"], "sdu" -> 
  SubsuperscriptBox[
   "\[SelectionPlaceholder]", "\[SelectionPlaceholder]", 
    "\[SelectionPlaceholder]"]},
FrontEndVersion->"11.0 for Microsoft Windows (64-bit) (September 21, 2016)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[580, 22, 150, 2, 101, "Title"],
Cell[CellGroupData[{
Cell[755, 28, 103, 1, 70, "Section"],
Cell[CellGroupData[{
Cell[883, 33, 256, 5, 46, "Input",
 InitializationCell->True],
Cell[CellGroupData[{
Cell[1164, 42, 2131, 35, 22, "Print"],
Cell[3298, 79, 2141, 36, 22, "Print"],
Cell[5442, 117, 2146, 36, 22, "Print"],
Cell[7591, 155, 2148, 36, 22, "Print"],
Cell[9742, 193, 2571, 41, 149, "Print"]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[12374, 241, 93, 1, 70, "Section"],
Cell[CellGroupData[{
Cell[12492, 246, 333, 6, 30, "ItemNumbered"],
Cell[12828, 254, 317, 5, 30, "ItemNumbered"],
Cell[13148, 261, 302, 4, 30, "ItemNumbered"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[13499, 271, 94, 1, 70, "Section"],
Cell[13596, 274, 1600, 47, 87, "Input",
 InitializationCell->True],
Cell[15199, 323, 8632, 203, 526, "Input",
 InitializationCell->True],
Cell[23834, 528, 1430, 37, 48, "Input",
 InitializationCell->True],
Cell[25267, 567, 802, 21, 48, "Input",
 InitializationCell->True],
Cell[26072, 590, 817, 18, 66, "Input",
 InitializationCell->True],
Cell[26892, 610, 4008, 83, 220, "Input",
 InitializationCell->True]
}, Open  ]],
Cell[CellGroupData[{
Cell[30937, 698, 115, 1, 70, "Section"],
Cell[31055, 701, 3934, 104, 194, "Input",
 InitializationCell->True],
Cell[CellGroupData[{
Cell[35014, 809, 99, 1, 49, "Subsection"],
Cell[35116, 812, 2369, 71, 145, "Input",
 InitializationCell->True],
Cell[CellGroupData[{
Cell[37510, 887, 684, 11, 30, "Input"],
Cell[38197, 900, 459, 11, 50, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[38693, 916, 544, 9, 30, "Input"],
Cell[39240, 927, 3408, 89, 103, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[42697, 1022, 127, 1, 49, "Subsection"],
Cell[CellGroupData[{
Cell[42849, 1027, 114, 1, 39, "Subsubsection"],
Cell[42966, 1030, 224, 5, 30, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[43227, 1040, 144, 2, 39, "Subsubsection"],
Cell[43374, 1044, 739, 14, 30, "Input"],
Cell[44116, 1060, 157, 4, 49, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[44310, 1069, 240, 3, 39, "Subsubsection"],
Cell[44553, 1074, 577, 12, 30, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[45167, 1091, 221, 3, 39, "Subsubsection"],
Cell[45391, 1096, 784, 15, 30, "Input"],
Cell[46178, 1113, 360, 9, 30, "Input"],
Cell[46541, 1124, 155, 3, 30, "Text"],
Cell[46699, 1129, 523, 9, 68, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[47259, 1143, 264, 3, 39, "Subsubsection"],
Cell[47526, 1148, 1424, 23, 30, "Input"],
Cell[48953, 1173, 380, 9, 30, "Input"],
Cell[49336, 1184, 202, 5, 49, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[49575, 1194, 181, 2, 39, "Subsubsection"],
Cell[CellGroupData[{
Cell[49781, 1200, 1845, 33, 69, "Input"],
Cell[51629, 1235, 3264, 89, 87, "Output"]
}, Open  ]],
Cell[54908, 1327, 219, 5, 49, "Text"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[55176, 1338, 98, 1, 49, "Subsection"],
Cell[55277, 1341, 1190, 33, 86, "Input",
 InitializationCell->True],
Cell[CellGroupData[{
Cell[56492, 1378, 697, 14, 30, "Input"],
Cell[57192, 1394, 8302, 221, 172, "Output"]
}, Open  ]],
Cell[65509, 1618, 501, 8, 68, "Text"]
}, Open  ]],
Cell[66025, 1629, 97, 1, 49, "Subsection"]
}, Open  ]]
}, Open  ]]
}
]
*)

